#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function enlarges an image by replicating bands, lines, and pixels. */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void DUPL (
IMAGE *in,      /*  I   Pointer to the input image.                           */
uint32_t ibgn,  /*  I   First replicated band.                                */
uint32_t jbgn,  /*  I   First replicated line.                                */
uint32_t kbgn,  /*  I   First replicated pixel.                               */
uint32_t ifact, /*  I   Band replication factor.                              */
uint32_t jfact, /*  I   Line replication factor.                              */
uint32_t kfact, /*  I   Pixel replication factor.                             */
uint32_t isize, /*  I   Number of replicated bands.                           */
uint32_t jsize, /*  I   Number of replicated lines.                           */
uint32_t ksize, /*  I   Number of replicated pixels.                          */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, l, m, n;
    char msg[SLEN];

    if (TIMES) TIMING(T_DUPL);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"DUPL");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (0 > ibgn  ||  ibgn+isize > in->nbnd) {
        MESSAGE('E'," Block band range to be duplicated must be located completely inside input image.");
        goto the_end;
    } else if (0 > jbgn  ||  jbgn+jsize > in->nlin) {
        MESSAGE('E'," Block line range to be duplicated must be located completely inside input image.");
        goto the_end;
    } else if (0 > kbgn  ||  kbgn+ksize > in->npix) {
        MESSAGE('E'," Block pixel range to be duplicated must be located completely inside input image.");
        goto the_end;
    } else if (ifact <= 0) {
        MESSAGE('E'," Band duplication factor must be greater than zero.");
        goto the_end;
    } else if (jfact <= 0) {
        MESSAGE('E'," Line duplication factor must be greater than zero.");
        goto the_end;
    } else if (kfact <= 0) {
        MESSAGE('E'," Pixel duplication factor must be greater than zero.");
        goto the_end;
    } else if (isize <= 0) {
        MESSAGE('E'," Size of block in bands to be duplicated must be greater then zero.");
        goto the_end;
    } else if (jsize <= 0) {
        MESSAGE('E'," Size of block in lines to be duplicated must be greater then zero.");
        goto the_end;
    } else if (ksize <= 0) {
        MESSAGE('E'," Size of block in pixels/line to be duplicated must be greater then zero.");
        goto the_end;
    } else if ((uint64_t)isize*(uint64_t)ifact > (uint64_t)UINT32_MAX) {
        MESSAGE('E'," Image size (number of bands) too large.");
        goto the_end;
    } else if ((uint64_t)jsize*(uint64_t)jfact > (uint64_t)UINT32_MAX) {
        MESSAGE('E'," Image size (number of lines) too large.");
        goto the_end;
    } else if ((uint64_t)ksize*(uint64_t)kfact > (uint64_t)UINT32_MAX) {
        MESSAGE('E'," Image size (number of pixels/line) too large.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                                   %s",in->text);
        MESSAGE('I',msg);
#ifndef SADIE_IPT
        sprintf(msg," First replicated band, line, pixel:            %lu, %lu, %lu",ibgn,jbgn,kbgn);
#else
        sprintf(msg," First replicated band, line, pixel:            %lu, %lu, %lu",ibgn+1,jbgn+1,kbgn+1);
#endif
        MESSAGE('I',msg);  
        sprintf(msg," Replication factor for bands, lines, pixels:   %lu, %lu, %lu",ifact,jfact,kfact);
        MESSAGE('I',msg);
        sprintf(msg," Number of replicated bands, lines, pixels:     %lu, %lu, %lu",isize,jsize,ksize);
        MESSAGE('I',msg); 
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(isize*ifact,jsize*jfact,ksize*kfact,out);
    if (!*out) goto the_end;

    /* duplicate the image data */
    for (i=ibgn; i<ibgn+isize; i++) {
        for (j=jbgn; j<jbgn+jsize; j++) {
            for (k=kbgn; k<kbgn+ksize; k++) {
                for (l=0; l<ifact; l++) {
                    for (m=0; m<jfact; m++) {
                        for (n=0; n<kfact; n++) {
                            (*out)->data[i*ifact+l][j*jfact+m][k*kfact+n] = in->data[i][j][k];
                        }
                    }
                }
            }
        }
    }
    (*out)->gmin = in->gmin;
    (*out)->gmax = in->gmax;

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
