#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This low-level function computes a discrete one-dimensional Fourier      */
/*   transform.                                                               */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void FFT1D (
short  option,  /*  I   Switch, indicating the direction of the transform:    */
                /*      FORWARD   -   forward Fourier transform is computed.  */
                /*      INVERSE   -   inverse Fourier transform is computed.  */
uint32_t size,  /*  I   Number of complex data points.                        */
double *sine,   /*  I   Array containing sine function values.                */
double *cosn,   /*  I   Array containing cosine function values.              */
double *array   /* I/O  Pointer to the complex data array[2*size].            */
/*----------------------------------------------------------------------------*/
) { register uint32_t n, m, l, k, j, i;
    register double temp, e_real, e_imag, o_real, o_imag;

    if (TIMES) TIMING(T_FFT1D);

    /* scramble the array */
    for (i=j=0; j<size; j+=1,i+=k) {
        if (i > j) {
            temp = array[2*i];
            array[2*i] = array[2*j];
            array[2*j] = temp;
            temp = array[2*i+1];
            array[2*i+1] = array[2*j+1];
            array[2*j+1] = temp;
        }
        for (k=size/2; 1<=k && k<=i; i-=k,k/=2) {
            continue;
        }
    }

    /* compute Fourier coefficients */
    for (i=2; i<=size; i*=2) {
        for (j=0; j<size/i; j++) {
            for (k=0; k<i/2; k++) {
                l = 2*(j*i+k);
                m = 2*(j*i+i/2+k);
                n = (uint32_t)((uint64_t)k*(uint64_t)size/(uint64_t)i);
                e_real = array[l  ];
                e_imag = array[l+1];
                o_real = array[m]*cosn[n] - array[m+1]*sine[n];
                o_imag = array[m]*sine[n] + array[m+1]*cosn[n];
                array[l  ] = (e_real+o_real)/2.0;
                array[l+1] = (e_imag+o_imag)/2.0;
                array[m  ] = (e_real-o_real)/2.0;
                array[m+1] = (e_imag-o_imag)/2.0;
            }
        }
    }

    if (option == INVERSE) {
        for (i=0; i<2*size; i++) {
            array[i] *= (double)size;
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the Fourier transform of an image.                */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   Bergland, G.D.:                                                          */
/*   "A guided tour of the fast Fourier transform."                           */
/*   IEEE Spectrum, Vol. 6, No. 7, July 1969                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void FFT2D (
IMAGE *in,      /*  I   Pointer to the input image.                           */
short option,   /*  I   Switch, indicating the direction of the transform:    */
                /*      FORWARD   -   forward Fourier transform is computed.  */
                /*      INVERSE   -   inverse Fourier transform is computed.  */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k;
    uint64_t total;
    char   msg[SLEN];
    double *sine=NULL, *cosn=NULL, *array=NULL, pinc, psum;
    PIXEL  temp;
    IMAGE  *img;

    if (TIMES) TIMING(T_FFT2D);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"FFT2D");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nlin != 1<<rnd(log10((double)in->nlin)/log10(2.0))) {
        MESSAGE('E'," Number of lines must be a power of two.");
        goto the_end;
    } else if (in->npix != 1<<rnd(log10((double)in->npix)/log10(2.0))) {
        MESSAGE('E'," Number of pixels/line must be a power of two.");
        goto the_end;
    } else if (option == FORWARD  &&  in->npix > UINT32_MAX/2) {
        MESSAGE('E'," Image size (number of pixels/line) too large.");
        goto the_end;
    } else if (option != FORWARD  &&  option != INVERSE) {
        MESSAGE('E'," Transformation option must be either FORWARD or INVERSE.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:   %s",in->text);
        MESSAGE('I',msg);
        if (option == FORWARD) {
            MESSAGE('I'," Forward FFT");
        } else if (option == INVERSE) {
            MESSAGE('I'," Inverse FFT");
        }
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,in->nlin,option==FORWARD ? 2*in->npix:in->npix/2,out);
    if (!*out) goto the_end;
    img = option == FORWARD  ?  *out : in;

    /* allocate and initialize trigonometric tables */
    i = max(img->nlin,img->npix/2);
    total = (uint64_t)i * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (sine=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)i * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (cosn=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    for (j=0; j<i; j++) {
        sine[j] = (double)option*sin((2.0*PI*(double)j)/(double)i);
        cosn[j] =                cos((2.0*PI*(double)j)/(double)i);
    }

    /* allocate array to hold one row/column of pixels */
    total = 2 * (uint64_t)i * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (array=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)img->nbnd/(double)(img->nlin+img->npix/2);

    /* Fourier transform bands */
    for (i=0; i<img->nbnd; i++) {

        /* convert from real to "complex" */
        if (option == FORWARD) {
            for (j=0; j<(*out)->nlin; j+=1) {
                for (k=0; k<(*out)->npix; k+=2) {
                    (*out)->data[i][j][k  ] = in->data[i][j][k/2];
                    (*out)->data[i][j][k+1] = (PIXEL)0;
                }
            }
        }

        /* fold */
        for (j=0; j<img->nlin/2; j++) {
            for (k=0; k<img->npix/2; k++) {
                temp = img->data[i][j][k];
                img->data[i][j][k] = img->data[i][img->nlin/2+j][img->npix/2+k];
                img->data[i][img->nlin/2+j][img->npix/2+k] = temp;
                temp = img->data[i][img->nlin/2+j][k];
                img->data[i][img->nlin/2+j][k] = img->data[i][j][img->npix/2+k];
                img->data[i][j][img->npix/2+k] = temp;
            }
        }

        /* Fourier transform rows */
        for (j=0; j<img->nlin; j++) {
            for (k=0; k<img->npix; k++) {
                array[k] = (double)img->data[i][j][k];
            }
            FFT1D(option,img->npix/2,sine,cosn,array);
            for (k=0; k<img->npix; k++) {
                img->data[i][j][k] = (PIXEL)array[k];
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }

        /* Fourier transform columns */
        for (k=0; k<img->npix; k+=2) {
            for (j=0; j<img->nlin; j+=1) {
                array[2*j  ] = (double)img->data[i][j][k  ];
                array[2*j+1] = (double)img->data[i][j][k+1];
            }
            FFT1D(option,img->nlin,sine,cosn,array);
            for (j=0; j<img->nlin; j+=1) {
                img->data[i][j][k  ] = (PIXEL)array[2*j  ];
                img->data[i][j][k+1] = (PIXEL)array[2*j+1];
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }

        /* unfold */
        for (j=0; j<img->nlin/2; j++) {
            for (k=0; k<img->npix/2; k++) {
                temp = img->data[i][j][k];
                img->data[i][j][k] = img->data[i][img->nlin/2+j][img->npix/2+k];
                img->data[i][img->nlin/2+j][img->npix/2+k] = temp;
                temp = img->data[i][img->nlin/2+j][k];
                img->data[i][img->nlin/2+j][k] = img->data[i][j][img->npix/2+k];
                img->data[i][j][img->npix/2+k] = temp;
            }
        }

        /* convert from "complex" to real */
        if (option == INVERSE) {
            for (j=0; j<in->nlin; j+=1) {
                for (k=0; k<in->npix; k+=2) {
                    (*out)->data[i][j][k/2] = in->data[i][j][k];
                }
            }
        }
    }

    the_end:
    if (sine)  free(sine);
    if (cosn)  free(cosn);
    if (array) free(array);
    if (TIMES) TIMING(T_EXIT);
}
