#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function spatially filters an image, using the Fast Fourier         */
/*   transform and a user-supplied frequency filter function. The filter      */
/*   must be real (amplitude only) and bilaterally symmetric.                 */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   General form of "flt":                                                   */
/*                                                                            */
/*      double flt (double fx, double fy, double xradius, double yradius)     */
/*      {                                                                     */
/*          return (filter amplitude at fx and fy);                           */
/*      }                                                                     */
/*                                                                            */
/*   where fx and fy are spatial frequencies in the range [0.0,0.5]           */
/*   cycles/pixel.  "Flt" is called by this function for all values of        */
/*   fx and fy and should return a filter function which is bilaterally       */
/*   symmetric about the fx and fy axes. "xradius" and "yradius" are the      */
/*   radii of the filter (in cycles/pixel) along the fx and fy frequency      */
/*   axes.  "Flt" may be made radially symmetric by setting "xradius"         */
/*   equal to "yradius".                                                      */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   Schowengerdt, R.A.:                                                      */
/*   "Techniques for Image Processing and Classification in Remote Sensing."  */
/*   Academic Press, New York, 1983, pp. 75-79                                */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void FILTER (
IMAGE  *in,     /*  I   Pointer to the input image.                           */
double (*flt)(double, double, double, double),
                /*  I   Pointer to a filter function.                         */
double xradius, /*  I   Radius of filter along fx in cycles/pixel.            */
double yradius, /*  I   Radius of filter along fy in cycles/line.             */
double option,  /*  I   Switch, indicating the type of filter:                */
                /*      = 0.0   -    low-pass   (e.g. filter = flt).          */
                /*      = 1.0   -    high-pass  (e.g. filter = 1.0-flt).      */
                /*      = 2.0   -    high-boost (e.g. filter = 2.0-flt).      */
                /*      else    -                     filter = option-flt.    */
IMAGE  **out    /*  O   Address of a pointer to the output image.             */
                /*      The input image may be overwritten (out = &in).       */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k;
    char   msg[SLEN];
    double fx, fy, ampl, pinc, psum;
    IMAGE  *scratch=NULL;

    if (TIMES) TIMING(T_FILTER);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"FILTER");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nlin != 1<<rnd(log10((double)in->nlin)/log10(2.0))) {
        MESSAGE('E'," Number of lines must be a power of two.");
        goto the_end;
    } else if (in->npix != 1<<rnd(log10((double)in->npix)/log10(2.0))) {
        MESSAGE('E'," Number of pixels/line must be a power of two.");
        goto the_end;
    } else if (!flt) {
        MESSAGE('E'," Can't identify filter function.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                   %s",in->text);
        MESSAGE('I',msg);
        if (flt == FLTBOX) {
            MESSAGE('I'," Box filter");
        } else if (flt == FLTCONE) {
            MESSAGE('I'," Cone filter");
        } else if (flt == FLTCYL) {
            MESSAGE('I'," Cylinder filter");
        } else if (flt == FLTEXP) {
            MESSAGE('I'," Exponential filter");
        } else if (flt == FLTGAUSS) {
            MESSAGE('I'," Gaussian filter");
        }
        sprintf(msg," Filter radius: cycles/line:  %8.4f",yradius);
        MESSAGE('I',msg);
        sprintf(msg,"                cycles/pixel: %8.4f",xradius);
        MESSAGE('I',msg);
        if (option == 0.0) {
            MESSAGE('I'," Low-pass filter");
        } else if (option == 1.0) {
            MESSAGE('I'," High-pass filter");
        } else {
            sprintf(msg," Parametric high-boost filter, c = %5.2f",option);
            MESSAGE('I',msg);
        }
        MESSAGE('I'," ...............");
    }

    /* Fourier transform image */
    FFT2D(in,FORWARD,&scratch);
    if (ABORT  ||  !scratch) goto the_end;
    sprintf(scratch->text,"Scratch");

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)scratch->nbnd/(double)(scratch->nlin/2-1);

    /* filter image */
    for (i=0; i<scratch->nbnd; i++) {
        for (j=1; j<scratch->nlin/2; j++) {
            fy = 0.5-(double)j/(double)scratch->nlin;
            for (k=2; k<scratch->npix/2-1; k+=2) {
                fx = 0.5-(double)k/(double)scratch->npix;
                ampl = (*flt)(fx,fy,xradius,yradius);
                if (option != 0.0) ampl = option - ampl;

                /* real components */
                scratch->data[i][j][k]               *= ampl;
                scratch->data[i][j][scratch->npix-k] *= ampl;

                /*  conjugate symmetry */
                scratch->data[i][scratch->nlin-j][k] = scratch->data[i][j][scratch->npix-k];
                scratch->data[i][scratch->nlin-j][scratch->npix-k] = scratch->data[i][j][k];

                /* imaginary components */
                scratch->data[i][j][k+1]               *= ampl;
                scratch->data[i][j][scratch->npix-k+1] *= ampl;

                /* conjugate symmetry */
                scratch->data[i][scratch->nlin-j][k+1] = -scratch->data[i][j][scratch->npix-k+1];
                scratch->data[i][scratch->nlin-j][scratch->npix-k+1] = -scratch->data[i][j][k+1];
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }
    }

    /* zero frequencies */
    for (i=0; i<scratch->nbnd; i++) {
        for (j=1; j<scratch->nlin; j++) {
            fx = 0.0;
            fy = 0.5 - (double)j/(double)scratch->nlin;
            ampl = (*flt)(fx,fy,xradius,yradius);
            if (option != 0.0) ampl = option - ampl;
            scratch->data[i][j][scratch->npix/2]   *= ampl;
            scratch->data[i][j][scratch->npix/2+1] *= ampl;
        }
    }
    for (i=0; i<scratch->nbnd; i++) {
        for (k=2; k<scratch->npix; k+=2) {
            fx = 0.5 - (double)k/(double)scratch->npix;
            fy = 0.0;
            ampl = (*flt)(fx,fy,xradius,yradius);
            if (option != 0.0) ampl = option - ampl;
            scratch->data[i][scratch->nlin/2][k]   *= ampl;
            scratch->data[i][scratch->nlin/2][k+1] *= ampl;
        }
    }

    /* folding frequencies */
    for (i=0; i<scratch->nbnd; i++) {
        for (j=0; j<scratch->nlin; j++) {
            fx = 0.5;
            fy = 0.5-(double)j/(double)scratch->nlin;
            ampl = (*flt)(fx,fy,xradius,yradius);
            if (option != 0.0) ampl = option - ampl;
            scratch->data[i][j][0] *= ampl;
            scratch->data[i][j][1] *= ampl;
        }
    }
    for (i=0; i<scratch->nbnd; i++) {
        for (k=2; k<scratch->npix; k+=2) {
            fx = 0.5-(double)k/(double)scratch->npix;
            fy = 0.5;
            ampl = (*flt)(fx,fy,xradius,yradius);
            if (option != 0.0) ampl = option - ampl;
            scratch->data[i][0][k]   *= ampl;
            scratch->data[i][0][k+1] *= ampl;
        }
    }

    /* Fourier transform image */
    FFT2D(scratch,INVERSE,out);

    the_end:
    if (scratch) RELMEM(scratch);
    if (TIMES)   TIMING(T_EXIT);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This library function computes a box spatial frequency filter            */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
double FLTBOX (
double fx,      /*  I   Spatial frequency.                                    */
double fy,      /*  I   Spatial frequency.                                    */
double xradius, /*  I   Radius of filter along fx.                            */
double yradius  /*  I   Radius of filter along fy.                            */
/*----------------------------------------------------------------------------*/
) { double ampl;

    if (TIMES) TIMING(T_FLTBOX);

    ampl = fabs(fx/xradius) <= 1.0  &&  fabs(fy/yradius) <= 1.0  ?  1.0 : 0.0;

    the_end:
    if (TIMES) TIMING(T_EXIT);
    return(ampl);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This library function computes a cone spatial frequency filter.          */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
double FLTCONE (
double fx,      /*  I   Spatial frequency.                                    */
double fy,      /*  I   Spatial frequency.                                    */
double xradius, /*  I   Radius of filter along fx.                            */
double yradius  /*  I   Radius of filter along fy.                            */
/*----------------------------------------------------------------------------*/
) { double ampl;

    if (TIMES) TIMING(T_FLTCONE);

    ampl = max(0.0,1.0-sqrt((fx*fx)/(xradius*xradius)+(fy*fy)/(yradius*yradius)));

    the_end:
    if (TIMES) TIMING(T_EXIT);
    return(ampl);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This library function computes a cylinder spatial frequency filter       */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
double FLTCYL (
double fx,      /*  I   Spatial frequency.                                    */
double fy,      /*  I   Spatial frequency.                                    */
double xradius, /*  I   Radius of filter along fx.                            */
double yradius  /*  I   Radius of filter along fy.                            */
/*----------------------------------------------------------------------------*/
) { double ampl;

    if (TIMES) TIMING(T_FLTCYL);

    ampl = sqrt((fx*fx)/(xradius*xradius)+(fy*fy)/(yradius*yradius)) <= 1.0  ?  1.0 : 0.0;

    the_end:
    if (TIMES) TIMING(T_EXIT);
    return(ampl);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This library function computes a negative exponential spatial frequency  */
/*   filter.                                                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
double FLTEXP (
double fx,      /*  I   Spatial frequency.                                    */
double fy,      /*  I   Spatial frequency.                                    */
double xradius, /*  I   Radius of filter along fx.                            */
double yradius  /*  I   Radius of filter along fy.                            */
/*----------------------------------------------------------------------------*/
) { double ampl;

    if (TIMES) TIMING(T_FLTEXP);

    ampl = exp(-sqrt((fx*fx)/(xradius*xradius)+(fy*fy)/(yradius*yradius)));

    the_end:
    if (TIMES) TIMING(T_EXIT);
    return(ampl);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This library function computes a gaussian spatial frequency filter.      */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
double FLTGAUSS (
double fx,      /*  I   Spatial frequency.                                    */
double fy,      /*  I   Spatial frequency.                                    */
double xradius, /*  I   Radius of filter along fx.                            */
double yradius  /*  I   Radius of filter along fy.                            */
/*----------------------------------------------------------------------------*/
) { double ampl;

    if (TIMES) TIMING(T_FLTGAUSS);

    ampl = exp(-((fx*fx)/(xradius*xradius)+(fy*fy)/(yradius*yradius)));

    the_end:
    if (TIMES) TIMING(T_EXIT);
    return(ampl);
}
