#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the modified Bessel Function of the first         */
/*   kind and zero order.  Taken from the Numerical Recipes book.             */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
static double mBesselF (
double x        /*  I   Argument to compute Io(x).                            */
/*----------------------------------------------------------------------------*/
) { double ax, bessel, y;

    if ((ax=fabs(x)) < 3.75) {
        y = x/3.75;
        y *= y;
        bessel = 1.0+y*(3.5156229+y*(3.0899424+y*(1.2067492+y*(0.2659732+y*(0.360768e-1+y*0.45813e-2)))));
    } else {
        y = 3.75/ax;
        bessel = (exp(ax)/sqrt(ax))*(0.39894228+y*(0.1328592e-1+y*(0.225319e-2+y*(-0.157565e-2+y*(0.916281e-2+y*(-0.2057706e-1+y*(0.2635537e-1+y*(-0.1647633e-1+y*0.392377e-2))))))));
    }
    return (bessel);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function creates an image as a two-dimensional test function.       */
/*   The function maximum graylevel is set to SMAX.                           */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void FUNCTION (
uint32_t nlin,  /*  I   Number of lines in the output image.                  */
uint32_t npix,  /*  I   Number of pixels/line in the output image.            */
double yradius, /*  I   Radius of the test function in y, in lines.           */
double xradius, /*  I   Radius of the test function in x, in pixels.          */
double alpha,   /*  I   Parameter for Kaiser window.                          */
short  option,  /*  I   Switch, indicating the type of the test function      */
                /*      ("r" is the distance of a pixel from the center):     */
                /*      CONE       -   SMAX * (1.0-r)                         */
                /*      NEGEXP     -   SMAX * exp(-r)                         */
                /*      GAUSS      -   SMAX * exp(-r**2)                      */
                /*      BOX        -   2*yradius lines by 2*xradius pixels    */
                /*      CYLINDER   -   r = 1.0                                */
                /*      HANNING    -                                          */
                /*      HAMMING    -                                          */
                /*      KAISER     -                                          */
                /*      SINC       -   sin(PI*r)/(PI*r)                       */
IMAGE  **out    /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t j, k; 
    uint64_t total;
    char   msg[SLEN];
    short  nlinrem, npixrem;
    double r, x, y, *elin=NULL, *epix=NULL;

    if (TIMES) TIMING(T_FUNCTION);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"FUNCTION");
        MESSAGE('I',"");
        sprintf(msg," Image size: lines:           %lu",nlin);
        MESSAGE('I',msg);
        sprintf(msg,"             pixels:          %lu",npix);
        MESSAGE('I',msg);
        sprintf(msg," Function radius: lines:    %12.4e",yradius);
        MESSAGE('I',msg);
        sprintf(msg,"                  pixels:   %12.4e",xradius);
        MESSAGE('I',msg);
        if (option == CONE) {
            MESSAGE('I'," Cone function");
        } else if (option == NEGEXP) {
            MESSAGE('I'," Negative exponential function");
        } else if (option == GAUSS) {
            MESSAGE('I'," Gaussian function");
        } else if (option == BOX) {
            MESSAGE('I'," Box function");
        } else if (option == CYLINDER) {
            MESSAGE('I'," Cylinder function");
        } else if (option == HANNING) {
            MESSAGE('I'," Hanning function");
        } else if (option == HAMMING) {
            MESSAGE('I'," Hamming function");
        } else if (option == KAISER) {
            sprintf(msg," Alpha:                     %12.4e",alpha);
            MESSAGE('I',msg);
            MESSAGE('I'," Kaiser function");
        } else if (option == SINC) {
            MESSAGE('I'," SINC function");
        }
        MESSAGE('I'," ...............");
    }

    /* check input */
    if (nlin <= 0) {
        MESSAGE('E'," Number of lines must be greater than zero.");
        goto the_end;
    } else if (npix <= 0) {
        MESSAGE('E'," Number of pixels/line must be greater than zero.");
        goto the_end;
    } else if (yradius <= 0.0) {
        MESSAGE('E'," Radius in lines of test function must be greater than zero.");
        goto the_end;
    } else if (xradius <= 0.0) {
        MESSAGE('E'," Radius in pixels/line of test function must be greater than zero.");
        goto the_end;
    } else if (option != CONE  &&  option != NEGEXP  &&  option != GAUSS  &&  option != BOX  &&  option != CYLINDER  &&  option != HANNING  &&  option != HAMMING  &&  option != KAISER  &&  option != SINC) {
        MESSAGE('E'," Option must be one of CONE, NEGEXP, GAUSS, BOX, CYLINDER, HANNING, HAMMING, KAISER, or SINC.");
        goto the_end;
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(1,nlin,npix,out);
    if (!*out) goto the_end;

    /* even/odd size parameter */
    nlinrem = nlin%2;
    npixrem = npix%2;

    /* set up the function image based on the option */
    switch (option) {
    case CONE:
        /* symmetric quadrants */
        for (j=1-nlinrem; j<nlin/2; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            for (k=1-npixrem; k<npix/2; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][j][k] =
                (*out)->data[0][j][npix-k-npixrem] =
                (*out)->data[0][nlin-j-nlinrem][k] =
                (*out)->data[0][nlin-j-nlinrem][npix-k-npixrem] =
                (PIXEL)max(0.0,(double)(SMAX-SMIN)*(1.0-sqrt(x+y)+(double)SMIN));
            }
        }
        /* zero coordinates */
        for (j=0; j<nlin; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            (*out)->data[0][j][npix/2] = (PIXEL)max(0.0,(double)(SMAX-SMIN)*(1.0-sqrt(y)+(double)SMIN));
        }
        for (k=0; k<npix; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            (*out)->data[0][nlin/2][k] = (PIXEL)max(0.0,(double)(SMAX-SMIN)*(1.0-sqrt(x)+(double)SMIN));
        }
        /* first column (if npix is even) */
        if (npixrem == 0) {
            x = ((double)(npix/2)*(double)(npix/2))/(xradius*xradius);
            for (j=0; j<nlin; j++) {
                y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
                (*out)->data[0][j][0] = (PIXEL)max(0.0,(double)(SMAX-SMIN)*(1.0-sqrt(x+y)+(double)SMIN));
            }
        }
        /* first row (if nlin is even) */
        if (nlinrem == 0) {
            y = ((double)(nlin/2)*(double)(nlin/2))/(yradius*yradius);
            for (k=0; k<npix; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][0][k] = (PIXEL)max(0.0,(double)(SMAX-SMIN)*(1.0-sqrt(x+y)+(double)SMIN));
            }
        }
        break;
    case NEGEXP:
        /* symmetric quadrants */
        for (j=1-nlinrem; j<nlin/2; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            for (k=1-npixrem; k<npix/2; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][j][k] =
                (*out)->data[0][j][npix-k-npixrem] =
                (*out)->data[0][nlin-j-nlinrem][k] =
                (*out)->data[0][nlin-j-nlinrem][npix-k-npixrem] =
                (PIXEL)((double)(SMAX-SMIN)*exp(-sqrt(x+y))+(double)SMIN);
            }
        }
        /* zero coordinates */
        for (j=0; j<nlin; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            (*out)->data[0][j][npix/2] = (PIXEL)((double)(SMAX-SMIN)*exp(-sqrt(y))+(double)SMIN);
        }
        for (k=0; k<npix; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            (*out)->data[0][nlin/2][k] = (PIXEL)((double)(SMAX-SMIN)*exp(-sqrt(x))+(double)SMIN);
        }
        /* first column (if npix is even) */
        if (npixrem == 0) {
            x = ((double)(npix/2)*(double)(npix/2))/(xradius*xradius);
            for (j=0; j<nlin; j++) {
                y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
                (*out)->data[0][j][0] = (PIXEL)((double)(SMAX-SMIN)*exp(-sqrt(x+y))+(double)SMIN);
            }
        }
        /* first row (if nlin is even) */
        if (nlinrem == 0) {
            y = ((double)(nlin/2)*(double)(nlin/2))/(yradius*yradius);
            for (k=0; k<npix; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][0][k] = (PIXEL)((double)(SMAX-SMIN)*exp(-sqrt(x+y))+(double)SMIN);
            }
        }
        break;
    case GAUSS:
        /* allocate memory for temporary storage */
        total = (uint64_t)(nlin/2 + nlinrem) * (uint64_t)sizeof(double);
        if (total > (uint64_t)SIZE_MAX  ||  (elin = (double *)malloc((size_t)total)) == NULL) {
            MESSAGE('E'," Memory request failed.");
            goto the_end;
        }
        total = (uint64_t)(npix/2 + npixrem) * (uint64_t)sizeof(double);
        if (total > (uint64_t)SIZE_MAX  ||  (epix = (double *)malloc((size_t)total)) == NULL) {
            MESSAGE('E'," Memory request failed.");
            goto the_end;
        }
        /* store one dimensional components */
        for (j=1-nlinrem; j<nlin/2; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            elin[j] = exp(-y);
        }
        for (k=1-npixrem; k<npix/2; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            epix[k] = exp(-x);
        }
        /* symmetric quadrants */
        for (j=1-nlinrem; j<nlin/2; j++) {
            for (k=1-npixrem; k<npix/2; k++) {
                (*out)->data[0][j][k] =
                (*out)->data[0][j][npix-k-npixrem] =
                (*out)->data[0][nlin-j-nlinrem][k] =
                (*out)->data[0][nlin-j-nlinrem][npix-k-npixrem] =
                (PIXEL)((double)(SMAX-SMIN)*(epix[k]*elin[j])+(double)SMIN);
            }
        }
        /* zero coordinates */
        for (j=0; j<nlin; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            (*out)->data[0][j][npix/2] = (PIXEL)((double)(SMAX-SMIN)*exp(-y)+(double)SMIN);
        }
        for (k=0; k<npix; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            (*out)->data[0][nlin/2][k] = (PIXEL)((double)(SMAX-SMIN)*exp(-x)+(double)SMIN);
        }
        /* first column (if npix is even) */
        if (npixrem == 0) {
            x = ((double)(npix/2)*(double)(npix/2))/(xradius*xradius);
            for (j=0; j<nlin; j++) {
                y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
                (*out)->data[0][j][0] = (PIXEL)((double)(SMAX-SMIN)*exp(-x-y)+(double)SMIN);
            }
        }
        /* first row (if nlin is even) */
        if (nlinrem == 0) {
            y = ((double)(nlin/2)*(double)(nlin/2))/(yradius*yradius);
            for (k=0; k<npix; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][0][k] = (PIXEL)((double)(SMAX-SMIN)*exp(-x-y)+(double)SMIN);
            }
        }
        break;
    case BOX:
        /* symmetric quadrants */
        for (j=1-nlinrem; j<nlin/2; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            for (k=1-npixrem; k<npix/2; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][j][k] =
                (*out)->data[0][j][npix-k-npixrem] =
                (*out)->data[0][nlin-j-nlinrem][k] =
                (*out)->data[0][nlin-j-nlinrem][npix-k-npixrem] =
                (PIXEL)(x <= 1.0  &&  y <= 1.0  ?  SMAX : SMIN);
            }
        }
        /* zero coordinates */
        for (j=0; j<nlin; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            (*out)->data[0][j][npix/2] = (PIXEL)(y <= 1.0  ?  SMAX : SMIN);
        }
        for (k=0; k<npix; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            (*out)->data[0][nlin/2][k] = (PIXEL)(x <= 1.0  ?  SMAX : SMIN);
        }
        /* first column (if npix is even) */
        if (npixrem == 0) {
            x = ((double)(npix/2)*(double)(npix/2))/(xradius*xradius);
            for (j=0; j<nlin; j++) {
                y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
                (*out)->data[0][j][0] = (PIXEL)(x <= 1.0  &&  y <= 1.0  ?  SMAX : SMIN);
            }
        }
        /* first row (if nlin is even) */
        if (nlinrem == 0) {
            y = ((double)(nlin/2)*(double)(nlin/2))/(yradius*yradius);
            for (k=0; k<npix; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][0][k] = (PIXEL)(x <= 1.0  &&  y <= 1.0  ?  SMAX : SMIN);
            }
        }
        break;
    case CYLINDER:
        /* symmetric quadrants */
        for (j=1-nlinrem; j<nlin/2; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            for (k=1-npixrem; k<npix/2; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][j][k] =
                (*out)->data[0][j][npix-k-npixrem] =
                (*out)->data[0][nlin-j-nlinrem][k] =
                (*out)->data[0][nlin-j-nlinrem][npix-k-npixrem] =
                (PIXEL)(sqrt(x+y) <= 1.0  ?  SMAX : SMIN);
            }
        }
        /* zero coordinates */
        for (j=0; j<nlin; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            (*out)->data[0][j][npix/2] = (PIXEL)(sqrt(y) <= 1.0  ?  SMAX : SMIN);
        }
        for (k=0; k<npix; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            (*out)->data[0][nlin/2][k] = (PIXEL)(sqrt(x) <= 1.0  ?  SMAX : SMIN);
        }
        /* first column (if npix is even) */
        if (npixrem == 0) {
            x = ((double)(npix/2)*(double)(npix/2))/(xradius*xradius);
            for (j=0; j<nlin; j++) {
                y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
                (*out)->data[0][j][0] = (PIXEL)(sqrt(x+y) <= 1.0  ?  SMAX : SMIN);
            }
        }
        /* first row (if nlin is even) */
        if (nlinrem == 0) {
            y = ((double)(nlin/2)*(double)(nlin/2))/(yradius*yradius);
            for (k=0; k<npix; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][0][k] = (PIXEL)(sqrt(x+y) <= 1.0  ?  SMAX : SMIN);
            }
        }
        break;
    case HANNING:
        /* symmetric quadrants */
        for (j=1-nlinrem; j<nlin/2; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            for (k=1-npixrem; k<npix/2; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][j][k] =
                (*out)->data[0][j][npix-k-npixrem] =
                (*out)->data[0][nlin-j-nlinrem][k] =
                (*out)->data[0][nlin-j-nlinrem][npix-k-npixrem] =
                (PIXEL)(sqrt(x+y) <= 1.0  ?  (SMAX-SMIN)*(0.50 + 0.50*cos(PI*sqrt(x+y))) + SMIN : SMIN);
            }
        }
        /* zero coordinates */
        for (j=0; j<nlin; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            (*out)->data[0][j][npix/2] = (PIXEL)(sqrt(y) <= 1.0  ?
						 (SMAX-SMIN)*(0.50 + 0.50*cos(PI*sqrt(y))) + SMIN : SMIN);
        }
        for (k=0; k<npix; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            (*out)->data[0][nlin/2][k] = (PIXEL)(sqrt(x) <= 1.0  ?
						 (SMAX-SMIN)*(0.50 + 0.50*cos(PI*sqrt(x))) + SMIN : SMIN);
        }
        /* first column (if npix is even) */
        if (npixrem == 0) {
            x = ((double)(npix/2)*(double)(npix/2))/(xradius*xradius);
            for (j=0; j<nlin; j++) {
                y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
                (*out)->data[0][j][0] = (PIXEL)(sqrt(x+y) <= 1.0  ?
						(SMAX-SMIN)*(0.50 + 0.50*cos(PI*sqrt(x+y))) + SMIN : SMIN);
            }
        }
        /* first row (if nlin is even) */
        if (nlinrem == 0) {
            y = ((double)(nlin/2)*(double)(nlin/2))/(yradius*yradius);
            for (k=0; k<npix; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][0][k] = (PIXEL)(sqrt(x+y) <= 1.0  ?
						(SMAX-SMIN)*(0.50 + 0.50*cos(PI*sqrt(x+y))) + SMIN : SMIN);
            }
        }
        break;
    case HAMMING:
        /* symmetric quadrants */
        for (j=1-nlinrem; j<nlin/2; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            for (k=1-npixrem; k<npix/2; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][j][k] =
                (*out)->data[0][j][npix-k-npixrem] =
                (*out)->data[0][nlin-j-nlinrem][k] =
                (*out)->data[0][nlin-j-nlinrem][npix-k-npixrem] =
                (PIXEL)(sqrt(x+y) <= 1.0  ?  (SMAX-SMIN)*(0.54 + 0.46*cos(PI*sqrt(x+y))) + SMIN : SMIN);
            }
        }
        /* zero coordinates */
        for (j=0; j<nlin; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            (*out)->data[0][j][npix/2] = (PIXEL)(sqrt(y) <= 1.0  ?
						 (SMAX-SMIN)*(0.54 + 0.46*cos(PI*sqrt(y))) + SMIN : SMIN);
        }
        for (k=0; k<npix; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            (*out)->data[0][nlin/2][k] = (PIXEL)(sqrt(x) <= 1.0  ?
						 (SMAX-SMIN)*(0.54 + 0.46*cos(PI*sqrt(x))) + SMIN : SMIN);
        }
        /* first column (if npix is even) */
        if (npixrem == 0) {
            x = ((double)(npix/2)*(double)(npix/2))/(xradius*xradius);
            for (j=0; j<nlin; j++) {
                y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
                (*out)->data[0][j][0] = (PIXEL)(sqrt(x+y) <= 1.0  ?
						(SMAX-SMIN)*(0.54 + 0.46*cos(PI*sqrt(x+y))) + SMIN : SMIN);
            }
        }
        /* first row (if nlin is even) */
        if (nlinrem == 0) {
            y = ((double)(nlin/2)*(double)(nlin/2))/(yradius*yradius);
            for (k=0; k<npix; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][0][k] = (PIXEL)(sqrt(x+y) <= 1.0  ?
						(SMAX-SMIN)*(0.54 + 0.46*cos(PI*sqrt(x+y))) + SMIN : SMIN);
            }
        }
        break;
    case KAISER:
        /* symmetric quadrants */
        for (j=1-nlinrem; j<nlin/2; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            for (k=1-npixrem; k<npix/2; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][j][k] =
                (*out)->data[0][j][npix-k-npixrem] =
                (*out)->data[0][nlin-j-nlinrem][k] =
                (*out)->data[0][nlin-j-nlinrem][npix-k-npixrem] =
                (PIXEL)(sqrt(x+y) <= 1.0  ?  (SMAX-SMIN)*(mBesselF(alpha*sqrt(1.0-(x+y)))/mBesselF(alpha)) + SMIN : SMIN);
            }
        }
        /* zero coordinates */
        for (j=0; j<nlin; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            (*out)->data[0][j][npix/2] = (PIXEL)(sqrt(y) <= 1.0  ?  (SMAX-SMIN)*(mBesselF(alpha*sqrt(1.0-y))/mBesselF(alpha)) + SMIN : SMIN);
        }
        for (k=0; k<npix; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            (*out)->data[0][nlin/2][k] = (PIXEL)(sqrt(x) <= 1.0  ?  (SMAX-SMIN)*(mBesselF(alpha*sqrt(1.0-x))/mBesselF(alpha)) + SMIN : SMIN);
        }
        /* first column (if npix is even) */
        if (npixrem == 0) {
            x = ((double)(npix/2)*(double)(npix/2))/(xradius*xradius);
            for (j=0; j<nlin; j++) {
                y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
                (*out)->data[0][j][0] = (PIXEL)(sqrt(x+y) <= 1.0  ?  (SMAX-SMIN)*(mBesselF(alpha*sqrt(1.0-(x+y)))/mBesselF(alpha)) + SMIN : SMIN);
            }
        }
        /* first row (if nlin is even) */
        if (nlinrem == 0) {
            y = ((double)(nlin/2)*(double)(nlin/2))/(yradius*yradius);
            for (k=0; k<npix; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                (*out)->data[0][0][k] = (PIXEL)(sqrt(x+y) <= 1.0  ?  (SMAX-SMIN)*(mBesselF(alpha*sqrt(1.0-(x+y)))/mBesselF(alpha)) + SMIN : SMIN);
            }
        }
        break;
    case SINC:
        /* symmetric quadrants */
        for (j=1-nlinrem; j<nlin/2; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            for (k=1-npixrem; k<npix/2; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                r = sqrt(x+y);
                (*out)->data[0][j][k] =
                (*out)->data[0][j][npix-k-npixrem] =
                (*out)->data[0][nlin-j-nlinrem][k] =
                (*out)->data[0][nlin-j-nlinrem][npix-k-npixrem] =
                (PIXEL)((double)(SMAX-SMIN)*(r == 0.0 ? 1.0 : sin(PI*r)/(PI*r))+(double)SMIN);
            }
        }
        /* zero coordinates */
        for (j=0; j<nlin; j++) {
            y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
            r = sqrt(y);
            (*out)->data[0][j][npix/2] = (PIXEL)((double)(SMAX-SMIN)*(r == 0.0 ? 1.0 : sin(PI*r)/(PI*r))+(double)SMIN);
        }
        for (k=0; k<npix; k++) {
            x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
            r = sqrt(x);
            (*out)->data[0][nlin/2][k] = (PIXEL)((double)(SMAX-SMIN)*(r == 0.0 ? 1.0 : sin(PI*r)/(PI*r))+(double)SMIN);
        }
        /* first column (if npix is even) */
        if (npixrem == 0) {
            x = ((double)(npix/2)*(double)(npix/2))/(xradius*xradius);
            for (j=0; j<nlin; j++) {
                y = (((double)j-(double)(nlin/2))*((double)j-(double)(nlin/2)))/(yradius*yradius);
                r = sqrt(x+y);
                (*out)->data[0][j][0] = (PIXEL)((double)(SMAX-SMIN)*(r == 0.0 ? 1.0 : sin(PI*r)/(PI*r))+(double)SMIN);
            }
        }
        /* first row (if nlin is even) */
        if (nlinrem == 0) {
            y = ((double)(nlin/2)*(double)(nlin/2))/(yradius*yradius);
            for (k=0; k<npix; k++) {
                x = (((double)k-(double)(npix/2))*((double)k-(double)(npix/2)))/(xradius*xradius);
                r = sqrt(x+y);
                (*out)->data[0][0][k] = (PIXEL)((double)(SMAX-SMIN)*(r == 0.0 ? 1.0 : sin(PI*r)/(PI*r))+(double)SMIN);
            }
        }
        break;
    }
    if (option != SINC) {
        (*out)->gmin = (*out)->data[0][0][0];
        (*out)->gmax = (*out)->data[0][nlin/2][npix/2];
    } else {
        RANGE(*out);
    }

    the_end:
    if (elin)  free(elin);
    if (epix)  free(epix);
    if (TIMES) TIMING(T_EXIT);
}
