#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1990 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This low-level function computes a cubic spline interpolation between    */
/*   four datapoints.                                                         */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
double CSPLINE (
double alpha,   /*  I   Interpolation function parameter (alpha < 0).         */
double dx,      /*  I   Fractional location of interpolated datapoint.        */
double g1,      /*  I   First datapoint.                                      */
double g2,      /*  I   Second datapoint.                                     */
double g3,      /*  I   Third datapoint.                                      */
double g4       /*  I   Fourth datapoint.                                     */
/*----------------------------------------------------------------------------*/
) { register double dx1, dx2, dx3, dx4, value;

    if (TIMES) TIMING(T_CSPLINE);

    /* compute interpolation value */
    dx1 = 1.0 + dx;
    dx2 =       dx;
    dx3 = 1.0 - dx;
    dx4 = 2.0 - dx;
    value = (-4.0 + 8.0*dx1 - 5.0*dx1*dx1 + dx1*dx1*dx1) * alpha  * g1  +
            (1.0 - (alpha+3.0)*dx2*dx2 + (alpha+2.0)*dx2*dx2*dx2) * g2  +
            (1.0 - (alpha+3.0)*dx3*dx3 + (alpha+2.0)*dx3*dx3*dx3) * g3  +
            (-4.0 + 8.0*dx4 - 5.0*dx4*dx4 + dx4*dx4*dx4) * alpha  * g4;

    the_end:
    if (TIMES) TIMING(T_EXIT);
    return(value);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function geometrically warps a "transformation" image with          */
/*   coordinate system (x',y') to a "reference" image with coordinate         */
/*   system (x,y), using a power series polynomial distortion model.          */
/*   The polynomial equations for the warp are:                               */
/*                                                                            */
/*      x  =    a1                      y  =    b1                            */
/*            + a2 * x'                       + b2 * x'                       */
/*            + a3 * y'                       + b3 * y'                       */
/*            + a4 * x' * y'                  + b4 * x' * y'                  */
/*            + a5 * x' * x'                  + b5 * x' * x'                  */
/*            + a6 * y' * y'                  + b6 * y' * y'                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void GEOMWARP (
IMAGE  *in,     /*  I   Pointer to the input image.                           */
uint32_t nbnd,  /*  I   Number of bands in the output image.                  */
uint32_t nlin,  /*  I   Number of lines in the output image.                  */
uint32_t npix,  /*  I   Number of pixels/line in the output image.            */
double *xc,     /*  I   Address of an array[nterms],                          */
                /*      containing the x-transformation coefficients.         */
double *yc,     /*  I   Address of an array[nterms],                          */
                /*      containing the y-transformation coefficients.         */
short  nterms,  /*  I   Number of terms in the power series polynomial.       */
double option,  /*  I   Switch, determines interpolation method:              */
                /*      < 0.0   -   Cubic interpolation (alpha = option).     */
                /*      = 0.0   -   Bilinear interpolation.                   */
                /*      > 0.0   -   Nearest neighbor interpolation.           */
PIXEL  glev,    /*  I   Graylevel to fill otherwise empty pixels.             */
IMAGE  **out    /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, ix, iy;
    char   msg[SLEN];
    double x, y, gl1, gl2, gl3, gl4, pinc, psum;

    if (TIMES) TIMING(T_GEOMWARP);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"GEOMWARP");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (nbnd > in->nbnd) {
        MESSAGE('E'," Number of bands in output image must be less than or equal to number of bands in input image.");
        goto the_end;
    } else if (MINTERMS > nterms) {
        sprintf(msg," Number of terms must be equal to or greater than %d.",MINTERMS);
        MESSAGE('E',msg);
        goto the_end;
    } else if (nterms > MAXTERMS) {
        sprintf(msg," Number of terms must be less than or equal to %d.",MAXTERMS);
        MESSAGE('E',msg);
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                  %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Output image size: bands:     %lu",nbnd);
        MESSAGE('I',msg);
        sprintf(msg,"                    lines:     %lu",nlin);
        MESSAGE('I',msg);
        sprintf(msg,"                    pixels:    %lu",npix);
        MESSAGE('I',msg);
        sprintf(msg," Fill graylevel:             %12.4e",glev);
        MESSAGE('I',msg);
        sprintf(msg," Number of polynomial terms:   %d",nterms);
        MESSAGE('I',msg);
        MESSAGE('I'," Transformation Coefficients:");
        sprintf(msg," Line  =   %12.4e                          Pixel  =   %12.4e",yc[0],xc[0]);
        MESSAGE('I',msg);
        sprintf(msg,"          +%12.4e * pixelref                         +%12.4e * pixelref",yc[1],xc[1]);
        MESSAGE('I',msg);
        sprintf(msg,"          +%12.4e * lineref                          +%12.4e * lineref",yc[2],xc[2]);
        MESSAGE('I',msg);
        if (nterms > 3) {
            sprintf(msg,"          +%12.4e * pixelref * lineref               +%12.4e * pixelref * lineref",yc[3],xc[3]);
            MESSAGE('I',msg);
        }
        if (nterms > 4) {
            sprintf(msg,"          +%12.4e * pixelref * pixelref              +%12.4e * pixelref * pixelref",yc[4],xc[4]);
            MESSAGE('I',msg);
        }
        if (nterms > 5) {
            sprintf(msg,"          +%12.4e * lineref  * lineref               +%12.4e * lineref  * lineref",yc[5],xc[5]);
            MESSAGE('I',msg);
        }
        if (option < 0.0) {
            sprintf(msg," Parametric cubic interpolation, alpha = %5.2f",option);
            MESSAGE('I',msg);
        } else if (option == 0.0) {
            MESSAGE('I'," Bilinear interpolation");
        } else if (option > 0.0) {
            MESSAGE('I'," Nearest-neighbor interpolation");
        }
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(nbnd,nlin,npix,out);
    if (!*out) goto the_end;

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)nbnd/(double)nlin;

    /* cubic interpolation */
    if (option < 0.0) {
        for (i=0; i<nbnd; i++) {
            for (j=0; j<nlin; j++) {
                for (k=0; k<npix; k++) {
                    x = y = 0.0;
                    switch (nterms) {
                    case 6:
                        x += xc[5]*(double)j*(double)j;
                        y += yc[5]*(double)j*(double)j;
                    case 5:
                        x += xc[4]*(double)k*(double)k;
                        y += yc[4]*(double)k*(double)k;
                    case 4:
                        x += xc[3]*(double)j*(double)k;
                        y += yc[3]*(double)j*(double)k;
                    case 3:
                        x += xc[0] + xc[1]*(double)k + xc[2]*(double)j;
                        y += yc[0] + yc[1]*(double)k + yc[2]*(double)j;
                    }
                    ix = (uint32_t)x;
                    iy = (uint32_t)y;
                    if (1 <= ix  &&  ix <= in->npix-3  &&  1 <= iy  &&  iy <= in->nlin-3) {
                        gl1 = CSPLINE(option,x-(double)ix,(double)in->data[i][iy-1][ix-1],(double)in->data[i][iy-1][ix],(double)in->data[i][iy-1][ix+1],(double)in->data[i][iy-1][ix+2]);
                        gl2 = CSPLINE(option,x-(double)ix,(double)in->data[i][iy  ][ix-1],(double)in->data[i][iy  ][ix],(double)in->data[i][iy  ][ix+1],(double)in->data[i][iy  ][ix+2]);
                        gl3 = CSPLINE(option,x-(double)ix,(double)in->data[i][iy+1][ix-1],(double)in->data[i][iy+1][ix],(double)in->data[i][iy+1][ix+1],(double)in->data[i][iy+1][ix+2]);
                        gl4 = CSPLINE(option,x-(double)ix,(double)in->data[i][iy+2][ix-1],(double)in->data[i][iy+2][ix],(double)in->data[i][iy+2][ix+1],(double)in->data[i][iy+2][ix+2]);
                        (*out)->data[i][j][k] = (PIXEL)CSPLINE(option,y-(double)iy,gl1,gl2,gl3,gl4);
                    } else if (ix == x  &&  1 <= ix  &&  ix <= in->npix-2  &&  1 <= iy  &&  iy <= in->nlin-3) {
                        (*out)->data[i][j][k] = (PIXEL)CSPLINE(option,y-(double)iy,(double)in->data[i][iy-1][ix],(double)in->data[i][iy][ix],(double)in->data[i][iy+1][ix],(double)in->data[i][iy+2][ix]);
                    } else if (1 <= ix  &&  ix <= in->npix-3  &&  iy == y  &&  1 <= iy  &&  iy <= in->nlin-2) {
                        (*out)->data[i][j][k] = (PIXEL)CSPLINE(option,x-(double)ix,(double)in->data[i][iy][ix-1],(double)in->data[i][iy][ix],(double)in->data[i][iy][ix+1],(double)in->data[i][iy][ix+2]);
                    } else if (ix == x  &&  1 <= ix  &&  ix <= in->npix-2  &&  iy == y  &&  1 <= iy  &&  iy <= in->nlin-2) {
                        (*out)->data[i][j][k] = in->data[i][iy][ix];
                    } else {
                        (*out)->data[i][j][k] = glev;
                    }
                }
                if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
            }
        }

    /* bilinear interpolation */
    } else if (option == 0.0) {
        for (i=0; i<nbnd; i++) {
            for (j=0; j<nlin; j++) {
                for (k=0; k<npix; k++) {
                    x = y = 0.0;
                    switch (nterms) {
                    case 6:
                        x += xc[5]*(double)j*(double)j;
                        y += yc[5]*(double)j*(double)j;
                    case 5:
                        x += xc[4]*(double)k*(double)k;
                        y += yc[4]*(double)k*(double)k;
                    case 4:
                        x += xc[3]*(double)j*(double)k;
                        y += yc[3]*(double)j*(double)k;
                    case 3:
                        x += xc[0] + xc[1]*(double)k + xc[2]*(double)j;
                        y += yc[0] + yc[1]*(double)k + yc[2]*(double)j;
                    }
                    ix = (uint32_t)x;
                    iy = (uint32_t)y;
                    if (0 <= ix  &&  ix <= in->npix-2  &&  0 <= iy  &&  iy <= in->nlin-2) {
                        gl1 = (double)in->data[i][iy  ][ix] + (x-(double)ix)*(double)(in->data[i][iy  ][ix+1]-in->data[i][iy  ][ix]);
                        gl2 = (double)in->data[i][iy+1][ix] + (x-(double)ix)*(double)(in->data[i][iy+1][ix+1]-in->data[i][iy+1][ix]);
                        (*out)->data[i][j][k] = (PIXEL)(gl1 + (y-(double)iy)*(gl2-gl1));
                    } else if (ix == x  &&  0 <= ix  &&  ix <= in->npix-1  &&  0 <= iy  &&  iy <= in->nlin-2) {
                        (*out)->data[i][j][k] = (PIXEL)((double)in->data[i][iy][ix] + (y-(double)iy)*(double)(in->data[i][iy+1][ix]-in->data[i][iy][ix]));
                    } else if (0 <= ix  &&  ix <= in->npix-2  &&  iy == y  &&  0 <= iy  &&  iy <= in->nlin-1) {
                        (*out)->data[i][j][k] = (PIXEL)((double)in->data[i][iy][ix] + (x-(double)ix)*(double)(in->data[i][iy][ix+1]-in->data[i][iy][ix]));
                    } else if (ix == x  &&  0 <= ix  &&  ix <= in->npix-1  &&  iy == y  &&  0 <= iy  &&  iy <= in->nlin-1) {
                        (*out)->data[i][j][k] = in->data[i][iy][ix];
                    } else {
                        (*out)->data[i][j][k] = glev;
                    }
                }
                if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
            }
        }

    /* nearest neighbor interpolation */
    } else /* (option > 0.0) */ {
        for (i=0; i<nbnd; i++) {
            for (j=0; j<nlin; j++) {
                for (k=0; k<npix; k++) {
                    x = y = 0.0;
                    switch (nterms) {
                    case 6:
                        x += xc[5]*(double)j*(double)j;
                        y += yc[5]*(double)j*(double)j;
                    case 5:
                        x += xc[4]*(double)k*(double)k;
                        y += yc[4]*(double)k*(double)k;
                    case 4:
                        x += xc[3]*(double)j*(double)k;
                        y += yc[3]*(double)j*(double)k;
                    case 3:
                        x += xc[0] + xc[1]*(double)k + xc[2]*(double)j;
                        y += yc[0] + yc[1]*(double)k + yc[2]*(double)j;
                    }
                    ix = rnd(x);
                    iy = rnd(y);
                    if (0 <= ix  &&  ix <= in->npix-1  &&  0 <= iy  &&  iy <= in->nlin-1) {
                        (*out)->data[i][j][k] = in->data[i][iy][ix];
                    } else {
                        (*out)->data[i][j][k] = glev;
                    }
                }
                if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
