#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function performs histogram equalization on an image.               */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void HISTEQ (
IMAGE *in,      /*  I   Pointer to the input image.                           */
uint32_t nlev,  /*  I   Histogram resolution                                  */
                /*      (and number of graylevels in the output image).       */
uint32_t inc,   /*  I   Line and pixel subsample increment.                   */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
                /*      The input image may be overwritten (out = &in).       */
/*----------------------------------------------------------------------------*/
) { register uint64_t i, j, k;
    uint64_t *hist=NULL, total;
    char msg[SLEN];
    double *dist=NULL, factor, width;

    if (TIMES) TIMING(T_HISTEQ);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"HISTEQ");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (inc <= 0) {
        MESSAGE('E'," Subsample increment must be greater than zero.");
        goto the_end;
    } else if (nlev <= 0) {
        MESSAGE('E'," Number of histogram bins must be greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                          %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Histogram resolution:                 %lu",nlev);
        MESSAGE('I',msg);
        sprintf(msg," Line and pixel subsample increment:   %lu",inc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,in->nlin,in->npix,out);
    if (!*out) goto the_end;

    /* allocate space for histogram and distribution arrays */
    total = (uint64_t)in->nbnd * (uint64_t)nlev * (uint64_t)sizeof(uint64_t);
    if (total > (uint64_t)SIZE_MAX  ||  (hist=(uint64_t *)malloc((size_t)total)) == NULL)  {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)in->nbnd * (uint64_t)nlev * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (dist=(double *)malloc((size_t)total)) == NULL)  {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* compute distribution function and scaling factor */
    RANGE(in);
    HISTOGRM(in,inc,IGNORE,in->gmin,in->gmax,nlev,hist);
    for (i=0; i<in->nbnd; i++) {
        for (j=1; j<nlev; j++) {
            hist[i*nlev+j] += hist[i*nlev+j-1];
        }
    }
    factor = (double)(in->gmax-in->gmin)/(double)hist[nlev-1];
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<nlev; j++) {
            dist[i*nlev+j] = hist[i*nlev+j]*factor;
        }
    }
    width  = (double)(in->gmax-in->gmin)/(double)(nlev-1);

    /* perform histogram equalization */
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<in->nlin; j++) {
            for (k=0; k<in->npix; k++) {
                (*out)->data[i][j][k] = (PIXEL)(dist[i*nlev+(uint64_t)((double)(in->data[i][j][k]-in->gmin)/width)] + (double)in->gmin);
            }
        }
    }
    (*out)->gmin = (PIXEL)((double)in->gmin+dist[0]);
    (*out)->gmax = (PIXEL)((double)in->gmin+dist[nlev-1]);

    the_end:
    if (hist)  free(hist);
    if (dist)  free(dist);
    if (TIMES) TIMING(T_EXIT);
}
