#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include "sadie.h"

#if defined SADIE_MACOS
#include <TextUtils.h>
#include <Files.h>
#endif

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function writes a SADIE image from main memory to disk.             */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void IMG2DISK (
IMAGE *in,           /*  I   Pointer to the image.                            */
unsigned char *name, /*  I   String, containing the name of the image file.   */
uint32_t nbit,       /*  I   Number of bits/pixel in image file.              */
short ref,           /*  I   Reference number for the image file.             */
short vref           /*  I   Volume reference number for the image file.      */
/*----------------------------------------------------------------------------*/
) { static char array[2] = { 0x00, 0x01 };
    register uint32_t i, j, k;
    uint64_t total;
    char msg[SLEN];
    long nbytes;
    IMAGE img;
#if defined SADIE_UNIX
    FILE *fp = NULL;
#elif defined SADIE_MACOS
    short error;
    Str255 pstr;
#endif

    if (TIMES) TIMING(T_IMG2DISK);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"IMG2DISK");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:               %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Output disk file:          %s",name);
        MESSAGE('I',msg);
        sprintf(msg," Output format:             0x%X, version %d",in->fileformat,in->fileversion);
        MESSAGE('I',msg);
        sprintf(msg," Number of bits/pixel/band: %lu",nbit);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* check number of bits per pixel */
    RANGE(in);
    if (nbit != SBIT   &&   (0ul > (long)in->gmin  ||  (long)in->gmax >= (1ul<<nbit)  ||  SBIT < nbit  ||  nbit > sizeof(long)*SBPB)) {
        sprintf(msg," Can't pack to %ld bits/pixel.",nbit);
        MESSAGE('E',msg);
        goto the_end;
    }
    img = *in;
    img.nbit = nbit;
    img.dataformat = (nbit == SBIT ? EXTDATA_IEEE_F32 : EXTDATA_BITSTRING);

    /* swap bytes, if necessary */
    if (*(short *)array != 1) {
        SWAP2((BYTE *)&img.fileformat);
        SWAP2((BYTE *)&img.fileversion);
        SWAP2((BYTE *)&img.application);
        SWAP2((BYTE *)&img.appversion);
        SWAP2((BYTE *)&img.dataformat);
        SWAP2((BYTE *)&img.datapacking);
        SWAP4((BYTE *)&img.npix);
        SWAP4((BYTE *)&img.nlin);
        SWAP4((BYTE *)&img.nbnd);
        SWAP4((BYTE *)&img.ntvl);
        SWAP4((BYTE *)&img.nbit);
        SWAP4((BYTE *)&img.gmin);
        SWAP4((BYTE *)&img.gmax);
    }

#if defined SADIE_UNIX
    /* open image file */
    if (!(fp=fopen(name,"w"))) {
        sprintf(msg," Can't open file %s.",name);
        MESSAGE('E',msg);
        goto the_end;
    }
#endif

    /* write image header to file */
    nbytes = (long)(total = (uint64_t)sizeof(img) - (uint64_t)sizeof(img.data));
#if defined SADIE_UNIX
    if (fwrite(&img,(size_t)total,1,fp) != 1) {
        sprintf(msg," Can't write image header to file %s.",name);
        MESSAGE('E',msg);
        goto the_end;
    }
#elif defined SADIE_MACOS
    if ((error=FSWrite(ref,&nbytes,&img)) != noErr) {
        sprintf(msg," Can't write image header to file %s.\rFile Manager Error Code: %d.",name,error);
        MESSAGE('E',msg);
        goto the_end;
    }
#endif

    /* write image data to file */
    nbytes = (long)(total = (uint64_t)ceil(((double)in->nbnd*(double)in->nlin*(double)in->npix*(double)nbit)/(double)SBPB));
    if (nbit < SBIT) {
        PACK(nbit,(uint64_t)in->nbnd*(uint64_t)in->nlin*(uint64_t)in->npix,in->data[0][0]);
    } else if (nbit == SBIT  &&  *(short *)array != 1) {
        for (i=0; i<in->nbnd; i++) {
            for (j=0; j<in->nlin; j++) {
                for (k=0; k<in->npix; k++) {
                    SWAP4((BYTE *)&in->data[i][j][k]);
                }
            }
        }
    }
#if defined SADIE_UNIX
    if (fwrite(in->data[0][0],(size_t)total,1,fp) != 1) {
        sprintf(msg," Can't write image data to file %s.",name);
        MESSAGE('E',msg);
    }
#elif defined SADIE_MACOS
    if ((error=FSWrite(ref,&nbytes,in->data[0][0])) != noErr) {
        sprintf(msg," Can't write image data to file %s.\rFile Manager Error Code: %d.",name,error);
        MESSAGE('E',msg);
        goto the_end;
    }
    c2pstrcpy(pstr,(char *)name);
    if ((error=FlushVol(pstr,vref)) != noErr  ||  (error=GetFPos(ref,&nbytes)) != noErr  ||  (error=SetEOF(ref,nbytes)) != noErr) {
        sprintf(msg," Can't write image data to file %s.\rFile Manager Error Code: %d.",name,error);
        MESSAGE('E',msg);
        goto the_end;
    }
#endif
    if (nbit < SBIT) {
        UNPACK(nbit,(uint64_t)in->nbnd*(uint64_t)in->nlin*(uint64_t)in->npix,in->data[0][0]);
    } else if (nbit == SBIT  &&  *(short *)array != 1) {
        for (i=0; i<in->nbnd; i++) {
            for (j=0; j<in->nlin; j++) {
                for (k=0; k<in->npix; k++) {
                    SWAP4((BYTE *)&in->data[i][j][k]);
                }
            }
        }
    }

    /* close image file */
    the_end:
#if defined SADIE_UNIX
    if (fp) {
        if (fclose(fp)) {
            sprintf(msg," Can't close file %s.",name);
            MESSAGE('W',msg);
        }
    }
#elif defined SADIE_MACOS
    if ((error=FSClose(ref)) != noErr) {
        sprintf(msg," Can't close file %s.\rFile Manager Error Code: %d.",name,error);
        MESSAGE('W',msg);
    }
#endif
    if (TIMES) TIMING(T_EXIT);
}
