#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include "sadie.h"

#if defined SADIE_MACOS
#include <Files.h>
#endif

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1990 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function imports an image from disk into main memory.               */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void IMPORT (
unsigned char *name, /*  I   String, containing the name of the image file.   */
short ref,           /*  I   Reference number of the image file.              */
short option,        /*  I   BIL, BSQ, or BIP.                                */
uint64_t size,       /*  I   Size, in bytes, of the image file.               */
uint64_t offset,     /*  I   Number of header bytes to be skipped.            */
uint32_t nbnd,       /*  I   Number of bands in the image file.               */
uint32_t nlin,       /*  I   Number of lines in the image file.               */
uint32_t npix,       /*  I   Number of pixels/line in the image file.         */
uint32_t nbit,       /*  I   Number of bits/pixel in the image file.          */
uint32_t ibgn,       /*  I   Index of the first band to be read.              */
uint32_t jbgn,       /*  I   Index of the first line to be read.              */
uint32_t kbgn,       /*  I   Index of the first pixel to be read.             */
uint32_t iinc,       /*  I   Band increment during sampling.                  */
uint32_t jinc,       /*  I   Line increment during sampling.                  */
uint32_t kinc,       /*  I   Pixels/line increment during sampling.           */
uint32_t iend,       /*  I   Index of the last band to be read.               */
uint32_t jend,       /*  I   Index of the last line to be read.               */
uint32_t kend,       /*  I   Index of the last pixel to be read.              */
IMAGE **out          /*  O   Address of a pointer to the output image.        */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, bands, lines, pixels; 
    uint64_t total;
    char msg[SLEN];
    long nbytes;
    PIXEL *line = NULL, *ptr;
#if defined SADIE_UNIX
    FILE *fp = NULL;
#elif defined SADIE_MACOS
    short error;
#endif

    if (TIMES) TIMING(T_IMPORT);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"IMPORT");
    }

    /* check input */
    if (nbnd <= 0) {
        MESSAGE('E'," Number of bands must be greater than zero.");
        goto the_end;
    } else if (nlin <= 0) {
        MESSAGE('E'," Number of lines must be greater than zero.");
        goto the_end;
    } else if (npix <= 0) {
        MESSAGE('E'," Number of pixels/line must be greater than zero.");
        goto the_end;
    } else if (nbit != SBIT   &&  (SBIT < nbit  ||  nbit > sizeof(long)*SBPB)) {
        sprintf(msg," Can't unpack from %lu bits/pixel.",nbit);
        MESSAGE('E',msg);
        goto the_end;
    } else if (0 > ibgn  ||  iend >= nbnd) {
        MESSAGE('E'," Image to be sampled must be located completely inside input image in the band dimension.");
        goto the_end;
    } else if (0 > jbgn  ||  jend >= nlin) {
        MESSAGE('E'," Image to be sampled must be located completely inside input image in the line dimension.");
        goto the_end;
    } else if (0 > kbgn  ||  kend >= npix) {
        MESSAGE('E'," Image to be sampled must be located completely inside input image in the pixel dimension.");
        goto the_end;
    } else if (iinc <= 0) {
        MESSAGE('E'," Band subsample increment must be greater than zero.");
        goto the_end;
    } else if (jinc <= 0) {
        MESSAGE('E'," Line subsample increment must be greater than zero.");
        goto the_end;
    } else if (kinc <= 0) {
        MESSAGE('E'," Pixel subsample increment must be greater than zero.");
        goto the_end;
    } else if (ibgn > iend) {
        MESSAGE('E'," Number of bands to be spanned must be greater than zero.");
        goto the_end;
    } else if (jbgn > jend) {
        MESSAGE('E'," Number of lines to be spanned must be greater than zero.");
        goto the_end;
    } else if (kbgn > kend) {
        MESSAGE('E'," Number of pixels/line to be spanned must be greater than zero.");
        goto the_end;
    }

    /* check that the number of bytes in the file matches the image size */
    nbytes = (long)(total = offset + (uint64_t)ceil(((double)nbnd*(double)nlin*(double)npix*(double)nbit)/(double)SBPB));
    if (size > 0  &&  size != total) {
        sprintf(msg," Size of file %s does not match image size:",name);
        MESSAGE('E',msg);
        sprintf(msg," File size = %llu, image size = %llu.",size,total);
        MESSAGE('E',msg);
        sprintf(msg," Nbnd = %lu, nlin = %lu, npix = %lu, nbit = %lu.",nbnd,nlin,npix,nbit);
        MESSAGE('E',msg);
        goto the_end;
    }

#if defined SADIE_UNIX
    /* open image file */
    if (!(fp=fopen(name,"r"))) {
        sprintf(msg," Can't open file %s.",name);
        MESSAGE('E',msg);
        goto the_end;
    }
#endif

    /* skip over the header block */
#if defined SADIE_UNIX
    if (fseek(fp,(long)offset,SEEK_SET) != 0) {
        sprintf(msg," Can't read header from file %s.",name);
        MESSAGE('E',msg);
        goto the_end;
    }
#elif defined SADIE_MACOS
    if ((error=SetFPos(ref,fsFromStart,(long)offset)) != noErr) {
        sprintf(msg," Can't read header from file %s.\rFile Manager Error Code: %d.",name,error);
        MESSAGE('E',msg);
        goto the_end;
    }
#endif

    /* create image of appropriate size */
    bands  = max(1,(iend-ibgn+1)/iinc);
    lines  = max(1,(jend-jbgn+1)/jinc);
    pixels = max(1,(kend-kbgn+1)/kinc);
    if (!CHECKIMG(*out)) GETMEM(bands,lines,pixels,out);
    if (!*out) goto the_end;

    /* read the image data from file */
    if (option == BIL) {
        nbytes = (long)(total = (uint64_t)npix * (uint64_t)sizeof(PIXEL));
        if (total > (uint64_t)SIZE_MAX  ||  (line=(PIXEL *)malloc((size_t)total)) == NULL) {
            MESSAGE('E'," Memory request failed.");
            goto the_end;
        }
        nbytes = (long)(total = (uint64_t)ceil((((double)npix*(double)nbit)/(double)SBPB)));
#if defined SADIE_UNIX
        if (fseek(fp,nbytes*(long)nbnd*(long)jbgn,SEEK_CUR) != 0) {
            sprintf(msg," Can't skip to line %d in file %s.",jbgn,name);
            MESSAGE('E',msg);
            goto the_end;
        }
#elif defined SADIE_MACOS
        if ((error=SetFPos(ref,fsFromMark,nbytes*(long)nbnd*(long)jbgn)) != noErr) {
            sprintf(msg," Can't skip to line %d in file %s.\rFile Manager Error Code: %d.",jbgn,name,error);
            MESSAGE('E',msg);
            goto the_end;
        }
#endif
        for (j = 0; j < lines; j++) {
#if defined SADIE_UNIX
            if (fseek(fp,nbytes*(long)nbnd*(long)(jinc-1),SEEK_CUR) != 0) {
                sprintf(msg," Can't line subsample by %d in file %s.",jinc,name);
                MESSAGE('E',msg);
                goto the_end;
            }
            if (fseek(fp,nbytes*(long)ibgn,SEEK_CUR) != 0) {
                sprintf(msg," Can't skip to band %d in file %s.",ibgn,name);
                MESSAGE('E',msg);
                goto the_end;
            }
#elif defined SADIE_MACOS
            if ((error=SetFPos(ref,fsFromMark,nbytes*(long)nbnd*(long)(jinc-1))) != noErr) {
                sprintf(msg," Can't line subsample by %d in file %s.\rFile Manager Error Code: %d.",jinc,name,error);
                MESSAGE('E',msg);
                goto the_end;
            }
            if ((error=SetFPos(ref,fsFromMark,nbytes*(long)ibgn)) != noErr) {
                sprintf(msg," Can't skip to band %d in file %s.\rFile Manager Error Code: %d.",ibgn,name,error);
                MESSAGE('E',msg);
                goto the_end;
            }
#endif
            for (i = 0; i < bands; i++) {
#if defined SADIE_UNIX
                if (fseek(fp,nbytes*(long)(iinc-1),SEEK_CUR) != 0) {
                    sprintf(msg," Can't band subsample by %d in file %s.",iinc,name);
                    MESSAGE('E',msg);
                    goto the_end;
                }
                if (fread(line,(size_t)total,1,fp) != 1) {
                    sprintf(msg," Can't read image data from file %s.",name);
                    MESSAGE('E',msg);
                    goto the_end;
                }
#elif defined SADIE_MACOS
                if ((error=SetFPos(ref,fsFromMark,nbytes*(long)(iinc-1))) != noErr) {
                    sprintf(msg," Can't band subsample by %d in file %s.\rFile Manager Error Code: %d.",iinc,name,error);
                    MESSAGE('E',msg);
                    goto the_end;
                }
                if ((error=FSRead(ref,&nbytes,line)) != noErr) {
                    sprintf(msg," Can't read image data from file %s.\rFile Manager Error Code: %d.",name,error);
                    MESSAGE('E',msg);
                    goto the_end;
                }
#endif
                if (nbit < SBIT) {
                    UNPACK(nbit,(uint64_t)npix,line);
                }
                for (k = 0, ptr = line + kbgn; k < pixels; k++, ptr += kinc) {
                    (*out)->data[i][j][k] = (PIXEL)*ptr;
                }
            }
#if defined SADIE_UNIX
            if (fseek(fp,nbytes*(long)(nbnd-iend-1),SEEK_CUR) != 0) {
                sprintf(msg," Can't skip to end of line in file %s.",name);
                MESSAGE('E',msg);
                goto the_end;
            }
#elif defined SADIE_MACOS
            if ((error=SetFPos(ref,fsFromMark,nbytes*(long)(nbnd-iend-1))) != noErr) {
                sprintf(msg," Can't skip to end of line in file %s.\rFile Manager Error Code: %d.",name,error);
                MESSAGE('E',msg);
                goto the_end;
            }
#endif
        }                
    } else if (option == BSQ) {
        nbytes = (long)(total = (uint64_t)npix * (uint64_t)sizeof(PIXEL));
        if (total > (uint64_t)SIZE_MAX  ||  (line=(PIXEL *)malloc((size_t)total)) == NULL) {
            MESSAGE('E'," Memory request failed.");
            goto the_end;
        }
        nbytes = (long)(total = (uint64_t)ceil((((double)npix*(double)nbit)/(double)SBPB)));
#if defined SADIE_UNIX
        if (fseek(fp,nbytes*(long)nlin*(long)ibgn,SEEK_CUR) != 0) {
            sprintf(msg," Can't skip to band %d in file %s.",ibgn,name);
            MESSAGE('E',msg);
            goto the_end;
        }
#elif defined SADIE_MACOS
        if ((error=SetFPos(ref,fsFromMark,nbytes*(long)nlin*(long)ibgn)) != noErr) {
            sprintf(msg," Can't skip to band %d in file %s.\rFile Manager Error Code: %d.",ibgn,name,error);
            MESSAGE('E',msg);
            goto the_end;
        }
#endif
        for (i = 0; i < bands; i++) {
#if defined SADIE_UNIX
            if (fseek(fp,nbytes*(long)nlin*(long)(iinc-1),SEEK_CUR) != 0) {
                sprintf(msg," Can't band subsample by %d in file %s.",iinc,name);
                MESSAGE('E',msg);
                goto the_end;
            }
            if (fseek(fp,nbytes*(long)jbgn,SEEK_CUR) != 0) {
                sprintf(msg," Can't skip to line %d in file %s.",jbgn,name);
                MESSAGE('E',msg);
                goto the_end;
            }
#elif defined SADIE_MACOS
            if ((error=SetFPos(ref,fsFromMark,nbytes*(long)nlin*(long)(iinc-1))) != noErr) {
                sprintf(msg," Can't band subsample by %d in file %s.\rFile Manager Error Code: %d.",iinc,name,error);
                MESSAGE('E',msg);
                goto the_end;
            }
            if ((error=SetFPos(ref,fsFromMark,nbytes*(long)jbgn)) != noErr) {
                sprintf(msg," Can't skip to line %d in file %s.\rFile Manager Error Code: %d.",jbgn,name,error);
                MESSAGE('E',msg);
                goto the_end;
            }
#endif
            for (j = 0; j < lines; j++) {
#if defined SADIE_UNIX
                if (fseek(fp,nbytes*(long)(jinc-1),SEEK_CUR) != 0) {
                    sprintf(msg," Can't line subsample by %d in file %s.",jinc,name);
                    MESSAGE('E',msg);
                    goto the_end;
                }
                if (fread(line,(size_t)total,1,fp) != 1) {
                    sprintf(msg," Can't read image data from file %s.",name);
                    MESSAGE('E',msg);
                    goto the_end;
                }
#elif defined SADIE_MACOS
                if ((error=SetFPos(ref,fsFromMark,nbytes*(long)(jinc-1))) != noErr) {
                    sprintf(msg," Can't line subsample by %d in file %s.\rFile Manager Error Code: %d.",jinc,name,error);
                    MESSAGE('E',msg);
                    goto the_end;
                }
                if ((error=FSRead(ref,&nbytes,line)) != noErr) {
                    sprintf(msg," Can't read image data from file %s.\rFile Manager Error Code: %d.",name,error);
                    MESSAGE('E',msg);
                    goto the_end;
                }
#endif
                if (nbit < SBIT) {
                    UNPACK(nbit,(uint64_t)npix,line);
                }
                for (k = 0, ptr = line + kbgn; k < pixels; k++, ptr += kinc) {
                    (*out)->data[i][j][k] = *ptr;
                }
            }
#if defined SADIE_UNIX
            if (fseek(fp,nbytes*(long)(nlin-jend-1),SEEK_CUR) != 0) {
                sprintf(msg," Can't skip to end of band in file %s.",name);
                MESSAGE('E',msg);
                goto the_end;
            }
#elif defined SADIE_MACOS
            if ((error=SetFPos(ref,fsFromMark,nbytes*(long)(nlin-jend-1))) != noErr) {
                sprintf(msg," Can't skip to end of band in file %s.\rFile Manager Error Code: %d.",name,error);
                MESSAGE('E',msg);
                goto the_end;
            }
#endif
        }                
    } else if (option == BIP) {
        nbytes = (long)(total = (uint64_t)nbnd * (uint64_t)sizeof(PIXEL));
        if (total > (uint64_t)SIZE_MAX  ||  (line=(PIXEL *)malloc((size_t)total)) == NULL) {
            MESSAGE('E'," Memory request failed.");
            goto the_end;
        }
        nbytes = (long)(total = (uint64_t)ceil((((double)nbnd*(double)nbit)/(double)SBPB)));
#if defined SADIE_UNIX
        if (fseek(fp,nbytes*(long)npix*(long)jbgn,SEEK_CUR) != 0) {
            sprintf(msg," Can't skip to line %d in file %s.",jbgn,name);
            MESSAGE('E',msg);
            goto the_end;
        }
#elif defined SADIE_MACOS
        if ((error=SetFPos(ref,fsFromMark,nbytes*(long)npix*(long)jbgn)) != noErr) {
            sprintf(msg," Can't skip to line %d in file %s.\rFile Manager Error Code: %d.",ibgn,name,error);
            MESSAGE('E',msg);
            goto the_end;
        }
#endif
        for (j = 0; j < lines; j++) {
#if defined SADIE_UNIX
            if (fseek(fp,nbytes*(long)npix*(long)(jinc-1),SEEK_CUR) != 0) {
                sprintf(msg," Can't line subsample by %d in file %s.",jinc,name);
                MESSAGE('E',msg);
                goto the_end;
            }
            if (fseek(fp,nbytes*(long)kbgn,SEEK_CUR) != 0) {
                sprintf(msg," Can't skip to pixel %d in file %s.",kbgn,name);
                MESSAGE('E',msg);
                goto the_end;
            }
#elif defined SADIE_MACOS
            if ((error=SetFPos(ref,fsFromMark,nbytes*(long)npix*(long)(jinc-1))) != noErr) {
                sprintf(msg," Can't line subsample by %d in file %s.\rFile Manager Error Code: %d.",iinc,name,error);
                MESSAGE('E',msg);
                goto the_end;
            }
            if ((error=SetFPos(ref,fsFromMark,nbytes*(long)kbgn)) != noErr) {
                sprintf(msg," Can't skip to pixel %d in file %s.\rFile Manager Error Code: %d.",jbgn,name,error);
                MESSAGE('E',msg);
                goto the_end;
            }
#endif
            for (k = 0; k < pixels; k++) {
#if defined SADIE_UNIX
                if (fseek(fp,nbytes*(long)(kinc-1),SEEK_CUR) != 0) {
                    sprintf(msg," Can't pixel subsample by %d in file %s.",kinc,name);
                    MESSAGE('E',msg);
                    goto the_end;
                }
                if (fread(line,(size_t)total,1,fp) != 1) {
                    sprintf(msg," Can't read image data from file %s.",name);
                    MESSAGE('E',msg);
                    goto the_end;
                }
#elif defined SADIE_MACOS
                if ((error=SetFPos(ref,fsFromMark,nbytes*(long)(kinc-1))) != noErr) {
                    sprintf(msg," Can't pixel subsample by %d in file %s.\rFile Manager Error Code: %d.",jinc,name,error);
                    MESSAGE('E',msg);
                    goto the_end;
                }
                if ((error=FSRead(ref,&nbytes,line)) != noErr) {
                    sprintf(msg," Can't read image data from file %s.\rFile Manager Error Code: %d.",name,error);
                    MESSAGE('E',msg);
                    goto the_end;
                }
#endif
                if (nbit < SBIT) {
                    UNPACK(nbit,(uint64_t)nbnd,line);
                }
                for (i = 0, ptr = line + ibgn; i < bands; i++, ptr += iinc) {
                    (*out)->data[i][j][k] = *ptr;
                }
            }
#if defined SADIE_UNIX
            if (fseek(fp,nbytes*(long)(npix-kend-1),SEEK_CUR) != 0) {
                sprintf(msg," Can't skip to end of line in file %s.",name);
                MESSAGE('E',msg);
                goto the_end;
            }
#elif defined SADIE_MACOS
            if ((error=SetFPos(ref,fsFromMark,nbytes*(long)(npix-kend-1))) != noErr) {
                sprintf(msg," Can't skip to end of line in file %s.\rFile Manager Error Code: %d.",name,error);
                MESSAGE('E',msg);
                goto the_end;
            }
#endif
        }
    }

    if (NAMES) {
        RANGE(*out);
        MESSAGE('I',"");
        sprintf(msg," Input file:                                 %s",name);
        MESSAGE('I',msg); 
        if (option == BIL) {
            MESSAGE('I'," Input format:                               Band-interleaved-by-line");
        } else if (option == BSQ) {
            MESSAGE('I'," Input format:                               Band-sequential");
        } else if (option == BIP) {      
            MESSAGE('I'," Input format:                               Band-interleaved-by-pixel");       
        }
        sprintf(msg," Header size:                                %llu",offset);
        MESSAGE('I',msg);
        sprintf(msg," Number of bands, lines, pixels:             %lu, %lu, %lu",nbnd,nlin,npix);
        MESSAGE('I',msg);
        sprintf(msg," Number of bits/pixel/band (image file):     %lu",nbit);
        MESSAGE('I',msg);
#ifndef SADIE_IPT
        sprintf(msg," Subsampled from band, line, pixel:          %lu, %lu, %lu",ibgn,jbgn,kbgn);
#else
        sprintf(msg," Subsampled from band, line, pixel:          %lu, %lu, %lu",ibgn+1,jbgn+1,kbgn+1);
#endif
        MESSAGE('I',msg);
#ifndef SADIE_IPT
        sprintf(msg," Subsampled to band, line, pixel:            %lu, %lu, %lu",iend,jend,kend);
#else
        sprintf(msg," Subsampled to band, line, pixel:            %lu, %lu, %lu",iend+1,jend+1,kend+1);
#endif
        MESSAGE('I',msg);
        sprintf(msg," Subsample increment: bands, lines, pixels:  %lu, %lu, %lu",iinc,jinc,kinc);
        MESSAGE('I',msg);
        sprintf(msg," Graylevel minimum, maximum:                 %10.4e, %10.4e",(*out)->gmin,(*out)->gmax);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* close image file */
    the_end:
#if defined SADIE_UNIX
    if (fp) {
        if (fclose(fp)) {
            sprintf(msg," Can't close file %s.",name);
            MESSAGE('W',msg);
        }
    }
#endif
    if (line)  free(line);
    if (TIMES) TIMING(T_EXIT);
}
