#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function assembles multiple images into a single composite mosaic.  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void MOSAIC (
IMAGE **in,     /*  I   Pointer to an array[nimg] containing pointers         */
                /*      to the input images.                                  */
uint32_t nimg,  /*  I   Number of input images.                               */
uint32_t nlin,  /*  I   Number of lines in the output image.                  */
uint32_t npix,  /*  I   Number of pixels/line in the output image.            */
uint32_t *jbgn, /*  I   Address of an array[nimg], containing the starting    */
                /*      line coordinates of the input images in the output.   */
uint32_t *kbgn, /*  I   Address of an array[nimg], containing the starting    */
                /*      pixel coordinates of the input images in the output.  */
PIXEL glev,     /*  I   Graylevel to fill otherwise empty pixels.             */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, l, nbnd, elin, epix;
    char msg[SLEN];

    if (TIMES) TIMING(T_MOSAIC);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"MOSAIC");
    }

    /* check input */
    for (i=0; i<nimg; i++) {
        if (!CHECKIMG(in[i])) {
            MESSAGE('E'," Can't identify input image(s).");
            goto the_end;
        }
    }
    for (nbnd=in[0]->nbnd,i=1; i<nimg; i++) {
        if (nbnd != in[i]->nbnd) {
            MESSAGE('E'," Number of bands must be identical in all images.");
            goto the_end;
        }
    }
    for (i=0; i<nimg; i++) {
        if (0 > jbgn[i]  ||  jbgn[i]+in[i]->nlin > nlin) {
            MESSAGE('E'," Input image(s) must be located completely inside output image in the line dimension.");
            goto the_end;
        } else if (0 > kbgn[i]  ||  kbgn[i]+in[i]->npix > npix) {
            MESSAGE('E'," Input image(s) must be located completely inside output image in the pixel dimension.");
            goto the_end;
        }
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Number of input images: %lu",nimg);
        MESSAGE('I',msg);
        sprintf(msg," Image size: lines:      %lu",nlin);
        MESSAGE('I',msg);
        sprintf(msg,"             pixels:     %lu",npix);
        MESSAGE('I',msg);
        for (i=0; i<nimg; i++) {
            sprintf(msg," Image:                  %s",in[i]->text);
            MESSAGE('I',msg);
            sprintf(msg," Offset: lines:          %lu",jbgn[i]);
            MESSAGE('I',msg);
            sprintf(msg,"         pixels:         %lu",kbgn[i]);
            MESSAGE('I',msg);
        }
        sprintf(msg," Fill graylevel:       %12.4e",glev);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(nbnd,nlin,npix,out);
    if (!*out) goto the_end;

    /* fill the background */
    for (i=0; i<nbnd; i++) {
        for (j=0; j<nlin; j++) {
            for (k=0; k<npix; k++) {
                (*out)->data[i][j][k] = glev;
            }
        }
    }

    /* assemble the mosaic */
    for (i=0; i<nbnd; i++) {
        for (l=0; l<nimg; l++) {
            elin = min(jbgn[l]+in[l]->nlin,nlin);
            epix = min(kbgn[l]+in[l]->npix,npix);
            for (j=jbgn[l]; j<elin; j++) {
                for (k=kbgn[l]; k<epix; k++) {
                    (*out)->data[i][j][k] = in[l]->data[i][j-jbgn[l]][k-kbgn[l]];
                }
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
