#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function performs a principal component transform on a multi-band   */
/*   image.                                                                   */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void PCT (
IMAGE *in,      /*  I   Pointer to the input image                            */
uint32_t inc,   /*  I   Line and pixel subsample increment to be used for     */
                /*      calculating the covariance matrix.                    */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
                /*      The input image may be overwritten (out = &in).       */
/*----------------------------------------------------------------------------*/
) { register uint64_t i, j, k, l, total;
    char   msg[SLEN];
    double *cov=NULL, *cor=NULL, *eig=NULL, sum, pinc, psum;

    if (TIMES) TIMING(T_PCT);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"PCT");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nbnd == 1) {
        MESSAGE('E'," Image must be multi-band.");
        goto the_end;
    } else if (inc <= 0) {
        MESSAGE('E'," Subsample increment must be greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                          %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Line and pixel subsample increment:   %lu",inc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,in->nlin,in->npix,out);
    if (!*out) goto the_end;

    /* allocate space for cov, cor, and eig */
    total = (uint64_t)in->nbnd * (uint64_t)in->nbnd * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (cov=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)in->nbnd * (uint64_t)in->nbnd * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (cor=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)in->nbnd * (uint64_t)in->nbnd * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (eig=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)in->nbnd/(double)in->nlin;

    /* compute covariance matrix of the input image and its eigenvectors */
    COVAR(in,inc,cov,cor);
    if (ABORT) goto the_end;
    EIGEN(cov,in->nbnd,eig);
    if (ABORT) goto the_end;

    /* perform principal component transform */
    for (i=0; i<in->nbnd; i++) {
        for (j=0; j<in->nlin; j++) {
            for (k=0; k<in->npix; k++) {
                for (sum=0.0,l=0; l<in->nbnd; l++) {
                    sum += (double)in->data[l][j][k]*eig[l*in->nbnd+i];
                }
                (*out)->data[i][j][k] = (PIXEL)sum;
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }
    }

    the_end:
    if (cov)   free(cov);
    if (cor)   free(cor);
    if (eig)   free(eig);
    if (TIMES) TIMING(T_EXIT);
}
