#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the power spectrum of an image, using the         */
/*   Fast Fourier transform.                                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void PSPECT (
IMAGE *in,      /*  I   Pointer to the input image.                           */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
                /*      The input image may be overwritten (out = &in).       */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k;
    double pinc, psum;
    IMAGE *scratch=NULL;
    char msg[SLEN];

    if (TIMES) TIMING(T_PSPECT);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"PSPECT");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (in->nlin != 1<<rnd(log10((double)in->nlin)/log10(2.0))) {
        MESSAGE('E'," Number of lines must be a power of two.");
        goto the_end;
    } else if (in->npix != 1<<rnd(log10((double)in->npix)/log10(2.0))) {
        MESSAGE('E'," Number of pixels/line must be a power of two.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:   %s",in->text);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* Fourier transform image */
    FFT2D(in,FORWARD,&scratch);
    if (ABORT  ||  !scratch) goto the_end;
    sprintf(scratch->text,"Scratch");

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,in->nlin,in->npix,out);
    if (!*out) goto the_end;

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)(*out)->nbnd/(double)(*out)->nlin;

    /* compute power spectrum */
    for (i=0; i<(*out)->nbnd; i++) {
        for (j=0; j<(*out)->nlin; j++) {
            for (k=0; k<(*out)->npix; k++) {
                (*out)->data[i][j][k] = (PIXEL)((double)scratch->data[i][j][2*k]*(double)scratch->data[i][j][2*k] + (double)scratch->data[i][j][2*k+1]*(double)scratch->data[i][j][2*k+1]);
            }
            if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
        }
    }

    the_end:
    if (scratch) RELMEM(scratch);
    if (TIMES)   TIMING(T_EXIT);
}

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the power spectrum of an image, using the         */
/*   Fast Fourier transform.                                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void PSPECT1D (
IMAGE *in,      /*  I   Pointer to the input image.                           */
short option,   /*  I   Switch to indicate type of power spectrum.            */
                /*      ROW_PSPECT - Average row power spectrum.              */
                /*      COL_PSPECT - Average column power spectrum.           */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
                /*      The input image may be overwritten (out = &in).       */
/*----------------------------------------------------------------------------*/
) { register uint64_t i, j, k;
    uint64_t total;
    char msg[SLEN];
    double *sine=NULL, *cosn=NULL, *array=NULL;
    IMAGE *img=NULL;

    if (TIMES) TIMING(T_PSPECT1D);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"PSPECT1D");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (option == COL_PSPECT  &&  in->nlin != 1<<rnd(log10((double)in->nlin)/log10(2.0))) {
        MESSAGE('E'," Number of lines must be a power of two.");
        goto the_end;
    } else if (option == ROW_PSPECT  &&  in->npix != 1<<rnd(log10((double)in->npix)/log10(2.0))) {
        MESSAGE('E'," Number of pixels/line must be a power of two.");
        goto the_end;
    } else if (option != ROW_PSPECT  &&  option != COL_PSPECT) {
        MESSAGE('E'," Option must be either ROW_PSPECT or COL_PSPECT.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:   %s",in->text);
        MESSAGE('I',msg);
        if (option == ROW_PSPECT) {
            MESSAGE('I'," Average Row Power Spectrum");
        } else if (option == COL_PSPECT) {
            MESSAGE('I'," Average Column Power Spectrum");
        }
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(in->nbnd,option == ROW_PSPECT ? 1 : in->nlin,option == ROW_PSPECT ? in->npix : 1,out);
    if (!*out) goto the_end;
    img = *out;

    /* allocate and initialize trigonometric tables */
    i = option == ROW_PSPECT ? img->npix : img->nlin;
    total = (uint64_t)i * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (sine=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)i * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (cosn=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    for (j=0; j<i; j++) {
        sine[j] = (double)FORWARD*sin((2.0*PI*(double)j)/(double)i);
        cosn[j] =                 cos((2.0*PI*(double)j)/(double)i);
    }
    
    /* allocate array to hold one row/column of pixels */
    total = (uint64_t)2 * (uint64_t)i * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (array=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }

    if (option == ROW_PSPECT) {
        
        for (i = 0; i < img->nbnd; i++) {

            /* initialize output to be used as accumulator */
            for (k = 0; k < img->npix; k++) {
                img->data[i][0][k] = (PIXEL)0;
            }
                
            for (j = 0; j < img->nlin; j++) {
        
                /* extract one row at a time */
                /* convert from real to "complex" and fold */
                for (k = 0; k < img->npix/2; k++) {
                    array[2*k  ] = in->data[i][j][img->npix/2+k];
                    array[2*k+1] = (PIXEL)0;
                    array[img->npix+2*k  ] = in->data[i][j][k];
                    array[img->npix+2*k+1] = (PIXEL)0;
                }
                
                /* Fourier transform rows */
                FFT1D(FORWARD,img->npix,sine,cosn,array);

                /* unfold and square */
                for (k = 0; k < img->npix/2; k++) {
                    img->data[i][0][k] += array[img->npix+2*k]*array[img->npix+2*k]+array[img->npix+2*k+1]*array[img->npix+2*k+1];
                    img->data[i][0][img->npix/2+k] += array[2*k]*array[2*k]+array[2*k+1]*array[2*k+1];
                }
            }

            /* average the line spectrum */
            for (k = 0; k < img->npix; k++) {
                img->data[i][0][k] /= (PIXEL)in->nlin;
            }
        }

    } else {

        for (i = 0; i < img->nbnd; i++) {

            /* initialize output to be used as accumulator */
            for (j = 0; j < img->nlin; j++) {
                img->data[i][j][0] = (PIXEL)0;
            }

            for (k = 0; k < img->npix; k++) {
                /* extract one column at a time */
                /* convert from real to "complex" and fold */
                for (j = 0; j < img->nlin/2; j++) {
                    array[2*j  ] = in->data[i][img->nlin/2+j][k];
                    array[2*j+1] = (PIXEL)0;
                    array[img->nlin+2*j  ] = in->data[i][j][k];
                    array[img->nlin+2*j+1] = (PIXEL)0;
                }

                /* Fourier transform columns */
                FFT1D(FORWARD,img->nlin,sine,cosn,array);

                /* unfold and square */
                for (j = 0; j < img->nlin/2; j++) {
                    img->data[i][j][0] += array[img->nlin+2*j]*array[img->nlin+2*j]+array[img->nlin+2*j+1]*array[img->nlin+2*j+1];
                    img->data[i][img->nlin/2+j][0] += array[2*j]*array[2*j]+array[2*j+1]*array[2*j+1];
                }

            }

            /* average the line spectrum */
            for (j = 0; j < img->nlin; j++) {
                img->data[i][j][0] /= (PIXEL)in->npix;
            }   
        }

    }

    the_end:
    if (sine)  free(sine);
    if (cosn)  free(cosn);
    if (array) free(array);
    if (TIMES) TIMING(T_EXIT);
}
