#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1990 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function takes as input a label map and two images, where the       */
/*   second image has twice the resolution of the label map and the first     */
/*   image. The output is a "refined" label map with the same resolution      */
/*   as the second image.                                                     */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   Mehldau, G.:                                                             */
/*   "An Image Segmentation Algorithm."                                       */
/*   University of Arizona, DIAL, Technical Report No. 90-3                   */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void REFINE (
IMAGE  *in[3],  /*  I   Array[3] of pointers to the input images.             */
                /*      in[0]:  Reduced label map.                            */
                /*      in[1]:  Reduced image.                                */
                /*      in[2]:  Refined image.                                */
IMAGE  **out    /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, l, m;
    char   msg[SLEN];
    double olddiff, newdiff, mean, pinc, psum;

    if (TIMES) TIMING(T_REFINE);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"REFINE");
    }

    /* check input */
    for (i=0; i<3; i++) {
        if (!CHECKIMG(in[i])) {
            MESSAGE('E'," Can't identify image.");
            goto the_end;
        }
    }
    if (in[0]->npix != in[1]->npix) {
        MESSAGE('E'," Number of pixels/line must be identical in reduced image and reduced label map.");
        goto the_end;
    } else if (in[0]->nlin != in[1]->nlin) {
        MESSAGE('E'," Number of lines must be identical in reduced image and reduced label map.");
        goto the_end;
    } else if (in[0]->nbnd != 1) {
        MESSAGE('E'," Reduced label map must be single-band.");
        goto the_end;
    } else if (2*in[1]->npix != in[2]->npix) {
        MESSAGE('E'," Number of pixels/line in reduced image must be half the number of pixels/line in regular image.");
        goto the_end;
    } else if (2*in[1]->nlin != in[2]->nlin) {
        MESSAGE('E'," Number of lines in reduced image must be half the number of lines in regular image.");
        goto the_end;
    } else if (in[1]->nbnd != in[2]->nbnd) {
        MESSAGE('E'," Number of bands must be identical in reduced image and regular image");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Reduced label map:   %s",in[0]->text);
        MESSAGE('I',msg);
        sprintf(msg," Reduced image:       %s",in[1]->text);
        MESSAGE('I',msg);
        sprintf(msg," Refined image:       %s",in[2]->text);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    if (!CHECKIMG(*out)) GETMEM(1,in[2]->nlin,in[2]->npix,out);
    if (!*out) goto the_end;

    /* initialize progress indicator */
    if (LINES  &&  !PROGRESS(psum=0.0)) goto the_end;
    pinc = 1.0/(double)(*out)->nlin;
  
    for (j=0; j<(*out)->nlin; j++) {
        for (k=0; k<(*out)->npix; k++) {
            switch (j%2 << 1  |  k%2) {

            /* type 1 */
            case 0:
                if (j > 0  &&  (*out)->data[0][j-1][k] == in[0]->data[0][j/2][k/2]  &&  (*out)->data[0][j-1][k] != in[0]->data[0][j/2-1][k/2]   ||
                    k > 0  &&  (*out)->data[0][j][k-1] == in[0]->data[0][j/2][k/2]  &&  (*out)->data[0][j][k-1] != in[0]->data[0][j/2][k/2-1]   ||   j == 0  &&  k == 0) {
                    (*out)->data[0][j][k] = in[0]->data[0][j/2][k/2];
                } else {
                    olddiff = -1.0;
                    if (j > 0) {
                        l = (j-1)/2;
                        m = k/2;
                        newdiff = 0.0;
                        for (i=0; i<in[1]->nbnd; i++) {
                            if (l < in[1]->nlin-1  &&  m > 0) {
                                mean = (double)(in[1]->data[i][l+1][m]+in[1]->data[i][l+1][m-1]+in[1]->data[i][l][m]+in[1]->data[i][l][m-1]) / 4.0;
                            } else if (l < in[1]->nlin-1) {
                                mean = (double)(in[1]->data[i][l+1][m]+in[1]->data[i][l][m]) / 2.0;
                            } else if (m > 0) {
                                mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l][m-1]) / 2.0;
                            } else {
                                mean = (double)(in[1]->data[i][l][m]);
                            }
                            newdiff += fabs((double)in[2]->data[i][j][k]-mean)*fabs((double)in[2]->data[i][j][k]-mean);
                        }
                        olddiff = newdiff;
                        (*out)->data[0][j][k] = (*out)->data[0][j-1][k];
                    }
                    if (k > 0) {
                        l = j/2;
                        m = (k-1)/2;
                        newdiff = 0.0;
                        for (i=0; i<in[1]->nbnd; i++) {
                            if (l > 0  &&  m < in[1]->npix-1) {
                                mean = (double)(in[1]->data[i][l][m+1]+in[1]->data[i][l][m]+in[1]->data[i][l-1][m+1]+in[1]->data[i][l-1][m]) / 4.0;
                            } else if (l > 0) {
                                mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l-1][m]) / 2.0;
                            } else if (m < in[1]->npix-1) {
                                mean = (double)(in[1]->data[i][l][m+1]+in[1]->data[i][l][m]) / 2.0;
                            } else {
                                mean = (double)(in[1]->data[i][l][m]);
                            }
                            newdiff += fabs((double)in[2]->data[i][j][k]-mean)*fabs((double)in[2]->data[i][j][k]-mean);
                        }
                        if (olddiff > newdiff  ||  olddiff < 0.0) {
                            olddiff = newdiff;
                            (*out)->data[0][j][k] = (*out)->data[0][j][k-1];
                        }
                    }
                    newdiff = 0.0;
                    for (i=0; i<in[1]->nbnd; i++) {
                        newdiff += fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2]))*fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2]));
                    }
                    if (olddiff > newdiff  ||  olddiff < 0.0) {
                        olddiff = newdiff;
                        (*out)->data[0][j][k] = in[0]->data[0][j/2][k/2];
                    }
                }
                break;

            /* type 2 */
            case 1:
                if (j > 0  &&  (*out)->data[0][j-1][k] == in[0]->data[0][j/2][k/2]  &&  (*out)->data[0][j-1][k] == in[0]->data[0][j/2-1][k/2]   ||  j == 0  &&  k == (*out)->npix-1) {
                    (*out)->data[0][j][k] = in[0]->data[0][j/2][k/2];
                } else {
                    olddiff = -1.0;
                    if (j > 0) {
                        l = (j-1)/2;
                        m = k/2;
                        newdiff = 0.0;
                        for (i=0; i<in[1]->nbnd; i++) {
                            if (l < in[0]->nlin-1  &&  m < in[0]->npix-1) {
                                mean = (double)(in[1]->data[i][l+1][m+1]+in[1]->data[i][l+1][m]+in[1]->data[i][l][m+1]+in[1]->data[i][l][m]) / 4.0;
                            } else if (l < in[0]->nlin-1) {
                                mean = (double)(in[1]->data[i][l+1][m]+in[1]->data[i][l][m]) / 2.0;
                            } else if (m < in[0]->npix-1) {
                                mean = (double)(in[1]->data[i][l][m+1]+in[1]->data[i][l][m]) / 2.0;
                            } else {
                                mean = (double)(in[1]->data[i][l][m]);
                            }
                            newdiff += fabs((double)in[2]->data[i][j][k]-mean)*fabs((double)in[2]->data[i][j][k]-mean);
                        }
                        olddiff = newdiff;
                        (*out)->data[0][j][k] = (*out)->data[0][j-1][k];
                    }
                    l = j/2;
                    m = (k-1)/2;
                    newdiff = 0.0;
                    for (i=0; i<in[1]->nbnd; i++) {
                        if (l > 0  &&  m > 0) {
                            mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l][m-1]+in[1]->data[i][l-1][m]+in[1]->data[i][l-1][m-1]) / 4.0;
                        } else if (l > 0) {
                            mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l-1][m]) / 2.0;
                        } else if (m > 0) {
                            mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l][m-1]) / 2.0;
                        } else {
                            mean = (double)(in[1]->data[i][l][m]);
                        }
                        newdiff += fabs((double)in[2]->data[i][j][k]-mean)*fabs((double)in[2]->data[i][j][k]-mean);
                    }
                    if (olddiff > newdiff  ||  olddiff < 0.0) {
                        olddiff = newdiff;
                        (*out)->data[0][j][k] = (*out)->data[0][j][k-1];
                    }
                    if (k/2 < in[1]->npix-1) {
                        newdiff = 0.0;
                        for (i=0; i<in[1]->nbnd; i++) {
                            newdiff += fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2+1]))*fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2+1]));
                        }
                        if (olddiff > newdiff  ||  olddiff < 0.0) {
                            olddiff = newdiff;
                            (*out)->data[0][j][k] = in[0]->data[0][j/2][k/2+1];
                        }
                    }
                    newdiff = 0.0;
                    for (i=0; i<in[1]->nbnd; i++) {
                        newdiff += fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2]))*fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2]));
                    }
                    if (olddiff > newdiff  ||  olddiff < 0.0) {
                        olddiff = newdiff;
                        (*out)->data[0][j][k] = in[0]->data[0][j/2][k/2];
                    }
                }
                break;

            /* type 3 */
            case 2:
                if (k > 0  &&  (*out)->data[0][j][k-1] == in[0]->data[0][j/2][k/2]  &&  (*out)->data[0][j][k-1] != in[0]->data[0][j/2][k/2-1]  ||  j == (*out)->nlin-1  &&  k == 0) {
                    (*out)->data[0][j][k] = in[0]->data[0][j/2][k/2];
                } else {
                    olddiff = -1.0;
                    l = (j-1)/2;
                    m = k/2;
                    newdiff = 0.0;
                    for (i=0; i<in[1]->nbnd; i++) {
                        if (l > 0  &&  m > 0) {
                            mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l][m-1]+in[1]->data[i][l-1][m]+in[1]->data[i][l-1][m-1]) / 4.0;
                        } else if (l > 0) {
                            mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l-1][m]) / 2.0;
                        } else if (m > 0) {
                            mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l][m-1]) / 2.0;
                        } else {
                            mean = (double)(in[1]->data[i][l][m]);
                        }
                        newdiff += fabs((double)in[2]->data[i][j][k]-mean)*fabs((double)in[2]->data[i][j][k]-mean);
                    }
                    olddiff = newdiff;
                    (*out)->data[0][j][k] = (*out)->data[0][j-1][k];
                    if (k > 0) {
                        l = j/2;
                        m = (k-1)/2;
                        newdiff = 0.0;
                        for (i=0; i<in[1]->nbnd; i++) {
                            if (l < in[0]->nlin-1  &&  m < in[0]->npix-1) {
                                mean = (double)(in[1]->data[i][l+1][m+1]+in[1]->data[i][l+1][m]+in[1]->data[i][l][m+1]+in[1]->data[i][l][m]) / 4.0;
                            } else if (l < in[0]->nlin-1) {
                                mean = (double)(in[1]->data[i][l+1][m]+in[1]->data[i][l][m]) / 2.0;
                            } else if (m < in[0]->npix-1) {
                                mean = (double)(in[1]->data[i][l][m+1]+in[1]->data[i][l][m]) / 2.0;
                            } else {
                                mean = (double)(in[1]->data[i][l][m]);
                            }
                            newdiff += fabs((double)in[2]->data[i][j][k]-mean)*fabs((double)in[2]->data[i][j][k]-mean);
                        }
                        if (olddiff > newdiff  ||  olddiff < 0.0) {
                            olddiff = newdiff;
                            (*out)->data[0][j][k] = (*out)->data[0][j][k-1];
                        }
                    }
                    newdiff = 0.0;
                    for (i=0; i<in[1]->nbnd; i++) {
                        newdiff += fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2]))*fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2]));
                    }
                    if (olddiff > newdiff  ||  olddiff < 0.0) {
                        olddiff = newdiff;
                        (*out)->data[0][j][k] = in[0]->data[0][j/2][k/2];
                    }
                }
                break;

            /* type 4 */
            case 3:
                if (j == (*out)->nlin-1  &&  k == (*out)->npix-1) {
                    (*out)->data[0][j][k] = in[0]->data[0][j/2][k/2];
                } else {
                    olddiff = -1.0;
                    l = (j-1)/2;
                    m = k/2;
                    newdiff = 0.0;
                    for (i=0; i<in[1]->nbnd; i++) {
                        if (l > 0  &&  m < in[1]->npix-1) {
                            mean = (double)(in[1]->data[i][l][m+1]+in[1]->data[i][l][m]+in[1]->data[i][l-1][m+1]+in[1]->data[i][l-1][m]) / 4.0;
                        } else if (l > 0) {
                            mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l-1][m]) / 2.0;
                        } else if (m < in[1]->npix-1) {
                            mean = (double)(in[1]->data[i][l][m+1]+in[1]->data[i][l][m]) / 2.0;
                        } else {
                            mean = (double)(in[1]->data[i][l][m]);
                        }
                        newdiff += fabs((double)in[2]->data[i][j][k]-mean)*fabs((double)in[2]->data[i][j][k]-mean);
                    }
                    olddiff = newdiff;
                    (*out)->data[0][j][k] = (*out)->data[0][j-1][k];
                    l = j/2;
                    m = (k-1)/2;
                    newdiff = 0.0;
                    for (i=0; i<in[1]->nbnd; i++) {
                        if (l < in[1]->nlin-1  &&  m > 0) {
                            mean = (double)(in[1]->data[i][l+1][m]+in[1]->data[i][l+1][m-1]+in[1]->data[i][l][m]+in[1]->data[i][l][m-1]) / 4.0;
                        } else if (l < in[1]->nlin-1) {
                            mean = (double)(in[1]->data[i][l+1][m]+in[1]->data[i][l][m]) / 2.0;
                        } else if (m > 0) {
                            mean = (double)(in[1]->data[i][l][m]+in[1]->data[i][l][m-1]) / 2.0;
                        } else {
                            mean = (double)(in[1]->data[i][l][m]);
                        }
                        newdiff += fabs((double)in[2]->data[i][j][k]-mean)*fabs((double)in[2]->data[i][j][k]-mean);
                    }
                    if (olddiff > newdiff  ||  olddiff < 0.0) {
                        olddiff = newdiff;
                        (*out)->data[0][j][k] = (*out)->data[0][j][k-1];
                    }
                    if (j/2 < in[1]->nlin-1) {
                        newdiff = 0.0;
                        for (i=0; i<in[1]->nbnd; i++) {
                            newdiff += fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2+1][k/2]))*fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2+1][k/2]));
                        }
                        if (olddiff > newdiff  ||  olddiff < 0.0) {
                            olddiff = newdiff;
                            (*out)->data[0][j][k] = in[0]->data[0][j/2+1][k/2];
                        }
                    }
                    if (k/2 < in[1]->npix-1) {
                        newdiff = 0.0;
                        for (i=0; i<in[1]->nbnd; i++) {
                            newdiff += fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2+1]))*fabs((double)(in[2]->data[i][j][k]-in[1]->data[i][j/2][k/2+1]));
                        }
                        if (olddiff > newdiff  ||  olddiff < 0.0) {
                            olddiff = newdiff;
                            (*out)->data[0][j][k] = in[0]->data[0][j/2][k/2+1];
                        }
                    }
                }
                break;

            }
        }
        if (LINES  &&  !PROGRESS(psum+=pinc)) goto the_end;
    }

    the_end:
    if (TIMES)   TIMING(T_EXIT);
}
