#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1993 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the minimum, maximum, mean, standard deviation,   */
/*   and covariance matrix from a set of graylevel values.                    */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void ROISTATS (
char   *name,   /*  I   Name of Region Of Interest.                           */
uint32_t nbnd,  /*  I   Number of bands.                                      */
uint64_t nval,  /*  I   Number of pixel values per band.                      */
PIXEL  *gval,   /*  I   Pointer to an array[nbnd*nval] of pixel values.       */
uint32_t inc,   /*  I   Pixel subsample increment.                            */
uint32_t nlev,  /*  I   > 0:   Number of bins in the histogram.               */
                /*      = 0:   Histogram is suppressed.                       */
PIXEL  *gmin,   /*  O   Pointer to an array[nbnd], containing the             */
                /*      minimum graylevel for each band.                      */
PIXEL  *gmax,   /*  O   Pointer to an array[nbnd], containing the             */
                /*      maximum graylevel for each band.                      */
double *mean,   /*  O   Pointer to an array[nbnd], containing the             */
                /*      mean graylevel for each band.                         */
double *dev,    /*  O   Pointer to an array[nbnd], containing the             */
                /*      graylevel standard deviation for each band.           */
uint64_t *zero, /*  O   Pointer to an array[nbnd], containing the             */
                /*      number of pixels with a zero graylevel for each band. */
double *cov,    /*  O   Pointer to an array[nbnd*nbnd], containing the        */
                /*      graylevel covariance matrix.                          */
double *cor     /*  O   Pointer to an array[nbnd*nbnd], containing the        */
                /*      graylevel correlation matrix.                         */
/*----------------------------------------------------------------------------*/
) { register uint64_t i, j, k;
    register PIXEL p, q;
    uint64_t total;
    char   msg[SLEN], minlbl[SLEN], maxlbl[SLEN];
    double factor, *sum=NULL, *sumsq=NULL;
    PIXEL  *phist=NULL;

    if (TIMES) TIMING(T_ROISTATS);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"ROISTATS");
    }

    /* check input */
    if (nbnd <= 0) {
        MESSAGE('E'," Number of bands must be greater than zero.");
        goto the_end;
    } else if (nval <= 0) {
        MESSAGE('E'," Number of values must be greater than or equal to zero.");
        goto the_end;
    } else if (inc <= 0) {
        MESSAGE('E'," Increment must be greater than zero.");
        goto the_end;
    } else if (nlev < 0) {
        MESSAGE('E'," Number of histogram bins must be greater than or equal to zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input region:               %s",name);
        MESSAGE('I',msg);
        sprintf(msg," Pixel subsample increment:  %lu",inc);
        MESSAGE('I',msg);
        if (nlev > 0) {
            sprintf(msg," Histogram resolution:       %lu",nlev);
            MESSAGE('I',msg);
        }
        MESSAGE('I'," ...............");
    }

    /* allocate space */
    total = (uint64_t)nbnd * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (sum=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)nbnd * (uint64_t)nbnd * (uint64_t)sizeof(double);
    if (total > (uint64_t)SIZE_MAX  ||  (sumsq=(double *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    if (nlev > 0) {
        total = (uint64_t)nbnd * (uint64_t)nlev * (uint64_t)sizeof(PIXEL);
        if (total > (uint64_t)SIZE_MAX  ||  (phist=(PIXEL *)malloc((size_t)total)) == NULL) {
            MESSAGE('E'," Memory request failed.");
            goto the_end;
        }
    }

    /* initialize summation arrays */
    for (i=0; i<nbnd; i++) {
        gmin[i] = gval[i*nval];
        gmax[i] = gval[i*nval];
        zero[i] = 0;
        sum[i]  = 0.0;
        for (j=0; j<nbnd; j++) {
            sumsq[i*nbnd+j] = 0.0;
        }
    }
    
    /* compute the sums */
    for (i=0; i<nbnd; i+=1) {
        for (j=0; j<nval; j+=inc) {
            p = gval[i*nval+j];
            gmin[i]          = min(gmin[i],p);
            gmax[i]          = max(gmax[i],p);
            zero[i]         += (uint64_t)(p == (PIXEL)0);
            sum[i]          += (double)p;
            sumsq[i*nbnd+i] += (double)(p*p);
        }
    }
    for (i=0; i<nbnd-1; i+=1) {
        for (j=i+1; j<nbnd; j+=1) {
            for (k=0; k<nval; k+=inc) {
                p = gval[i*nval+k];
                q = gval[j*nval+k];
                sumsq[i*nbnd+j] = sumsq[j*nbnd+i] += (double)(p*q);
            }
        }
    }

    /* calculate the statistics */
    for (i=0; i<nbnd; i++) {
        mean[i] = sum[i]/(double)(nval/inc+(nval%inc ? 1:0));
        dev[i]  = sqrt(sumsq[i*nbnd+i]/(double)(nval/inc+(nval%inc ? 1:0)) - mean[i]*mean[i]);
    }

    /* output the statistics */
    MESSAGE('I',"");
    MESSAGE('I'," Band     Minimum     Maximum        Mean   Deviation       Total      Zero/Non-Zero");
    for (i=0; i<nbnd; i++) {
        sprintf(msg,"%5llu%12.4e%12.4e%12.4e%12.4e%12llu%10llu/%llu",i+1,(double)gmin[i],(double)gmax[i],mean[i],dev[i],nval,zero[i],nval-zero[i]);
        MESSAGE('I',msg);
    }

    if (nlev > 0) {

        /* initialize */
        for (i=0; i<nbnd*nlev; i++) {
            phist[i] = (PIXEL)0;
        }
        p = gmin[0];
        q = gmax[0];
        for (i=1; i<nbnd; i++) {
            p = min(p,gmin[i]);
            q = max(q,gmax[i]);
        }
        factor = (double)nlev/(double)(q-p+1);

        /* compute histogram(s) */
        for (i=0; i<nbnd; i++) {
            for (j=0; j<nval; j+=inc) {
                if (gval[i*nval+j] <= p) {
                    phist[i*nlev+0] += (PIXEL)1;
                } else if (p < gval[i*nval+j]  &&  gval[i*nval+j] < q) {
                    phist[i*nlev+(uint64_t)((double)(gval[i*nval+j]-p)*factor)] += (PIXEL)1;
                } else {
                    phist[i*nlev+nlev-1] += (PIXEL)1;
                }
            }
        }

        /* output histogram(s) */
        MESSAGE('I',"");
        sprintf(msg," Histogram has %lu levels between %.4e and %.4e",nlev,p,q);
        MESSAGE('I',msg);
        sprintf(minlbl,"%.2f",p);
        sprintf(maxlbl,"%.2f",q);
        PLOT(phist,nbnd,nlev,minlbl,maxlbl,NORM);
    }

    if (cov != NULL) {

        /* calculate covariance matrix */
        for (i=0; i<nbnd; i++) {
            for (j=0; j<nbnd; j++) {
                cov[i*nbnd+j] = (sumsq[i*nbnd+j]-sum[i]*sum[j]/(double)(nval/inc+(nval%inc ? 1:0)))/((double)(nval/inc+(nval%inc ? 1:0)-1));
            }
        }

        /* output covariance matrix */
        MESSAGE('I',"");
        MESSAGE('I'," Covariance Matrix");
        for (i=0; i<nbnd; i++) {
            sprintf(msg," ");
            for (j=0; j<min(3,nbnd); j++) {
                sprintf(msg+strlen(msg),"%12.4e",cov[i*nbnd+j]);
            }
            MESSAGE('I',msg);
        }
    }

    if (cov != NULL  &&  cor != NULL) {

        /* calculate correlation matrix */
        for (i=0; i<nbnd; i++) {
            for (j=0; j<nbnd; j++) {
                cor[i*nbnd+j] = cov[i*nbnd+j]/sqrt(cov[i*nbnd+i]*cov[j*nbnd+j]);
            }
        }

        /* output correlation matrix */
        MESSAGE('I',"");
        MESSAGE('I'," Correlation Matrix");
        for (i=0; i<nbnd; i++) {
            sprintf(msg," ");
            for (j=0; j<min(3,nbnd); j++) {
                sprintf(msg+strlen(msg),"%12.4e",cor[i*nbnd+j]);
            }
            MESSAGE('I',msg);
        }
    }

    the_end:
    if (sum)   free(sum);
    if (sumsq) free(sumsq);
    if (phist) free(phist);
    if (TIMES) TIMING(T_EXIT);
}
