#ifndef SADIE_H
#define SADIE_H

#ifdef __cplusplus
extern "C" {                            /* Enable the SADIE library to be linked to C++ code */
#endif

/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <limits.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <setjmp.h>

/*----------------------------------------------------------------------------*/

#if defined SADIE_UNIX
#define SYSTEM "UNIX (Solaris/Linux/MacOSX)"
#elif defined SADIE_MACOS
#define SYSTEM "MacOS"
#endif

#if defined SADIE_LIB
#ifndef HAVE_UINT64_T
#if SIZEOF_LONG == 8
typedef long int64_t;
typedef unsigned long uint64_t;
#elif !defined _NO_LONGLONG
typedef long long int64_t;
typedef unsigned long long uint64_t;
#endif /* (SIZEOF_LONG != 8) && (!defined _NO_LONGLONG) */
#endif /* !defined HAVE_UINT64_T */
#define LOCALE "libsadie"
#elif defined SADIE_IPT
#define LOCALE "IPT"
#else
#define LOCALE "SADIE"
#endif

#define DIR     ""                      /* Location of source code */

#define DBND    65535                   /* Number of bands available on the display */
#define DLIN    65535                   /* Number of lines available on the display */
#define DPIX    65535                   /* Number of pixels available on the display */
#define DBIT    8                       /* Number of bits per band */
#define DMIN    128.5                   /* Minimum color table entry on the display */
#define DMAX    255.5                   /* Maximum color table entry on the display */

#define SBPB    8                       /* System value for number of bits/byte */
#define SBIT    sizeof(PIXEL)*SBPB      /* System default for number of bits/pixel */

#define SMIN    0                       /* System default for minimum gray level */
#define SMAX    255                     /* System default for maximum gray level */

#define SLEN    1024                    /* System default for length of strings */

#define TLEN                468         /* Length of text string in IMAGE data structure */
                                        /* Adjusted, so that the overall size of an IMAGE */
                                        /* (without the data) equals 512 Bytes */

#define FILE_FORMAT         0xAD1E
#define FILE_VERSION        0x0001

#define APP_SADIE           0
#define APP_IPT             1
#define APP_TCLSADIE        2
#define APP_TREES           3

#define EXTDATA_BITSTRING   0x0000
#define EXTDATA_UINT8       0x0001
#define EXTDATA_UINT16      0x0002
#define EXTDATA_UINT24      0x0003
#define EXTDATA_UINT32      0x0004
#define EXTDATA_IEEE_F32    0x0005

#define BE_BITPACKED        0x0000
#define BE_BYTEPACKED       0x0001
#define LE_BYTEPACKED       0x0002

#define PIXEL_DIMENSION     0x01
#define LINE_DIMENSION      0x02
#define BAND_DIMENSION      0x04
#define INTERVAL_DIMENSION  0x08

#define VALUES_NORMAL       0x00
#define VALUES_INVERTED     0x80

/*----------------------------------------------------------------------------*/

typedef unsigned char   BYTE;       /* Data type of unit size */

typedef float           PIXEL;      /* Pixel data type */

typedef struct {                    /* Image data structure */
    uint16_t   fileformat;          /* File format (magic number) */
    uint16_t   fileversion;         /* File format version */
    uint16_t   application;         /* Application identifier */
    uint16_t   appversion;          /* Application version */
    uint16_t   dataformat;          /* Data format */
    uint16_t   datapacking;         /* Data packing */
    uint8_t    datarotation[4];     /* Data rotation */
    uint32_t   npix;                /* Number of pixels/line */
    uint32_t   nlin;                /* Number of lines */
    uint32_t   nbnd;                /* Number of bands */
    uint32_t   ntvl;                /* Number of intervals (reserved for future use) */
    uint32_t   nbit;                /* Number of bits/pixel (for disk I/O) */
    PIXEL      gmin;                /* Minimum gray level in image */
    PIXEL      gmax;                /* Maximum gray level in image */
    BYTE       text[TLEN];          /* Application-specific data */
    PIXEL      ***data;             /* Pointer to the data itself */
}                       IMAGE;

typedef struct {                    /* Old (v0) image data structure */
    short  nbnd;                    /* Number of bands */
    short  nlin;                    /* Number of lines */
    short  npix;                    /* Number of pixels/line */
    short  nbit;                    /* Number of bits/pixel (for disk I/O) */
    PIXEL  gmin;                    /* Minimum gray level in image */
    PIXEL  gmax;                    /* Maximum gray level in image */
    char   pstr[1];                 /* Length byte for Pascal strings */
    char   text[495];               /* Image/window name */
    PIXEL  ***data;                 /* Pointer to the data itself */
}                       IMAGEv0;

typedef struct ROI {                    /* Region of Interest data structure */
        unsigned long   tmpSelect;
        unsigned long   select;
        char            *name;
        uint64_t        nval;
        uint64_t        *zero;
        PIXEL           *gmin;
        PIXEL           *gmax;
        double          *mean;
        double          *dev;
        double          *cov;
        double          *cor;
        struct ROI      *next;
}                       ROI, *ROIPtr;

typedef enum {                          /* Definitions for timing function */
    T_EXIT,
    T_BILTOBSQ,
    T_BITPLANE,
    T_BSQTOBIL,
    T_CHECKER,
    T_CHECKIMG,
    T_CHIRP,
    T_CLUSTER,
    T_COM2REAL,
    T_COMBINE,
    T_COMPARE,
    T_CONTNORM,
    T_CONTOUR,
    T_COVAR,
    T_CSPLINE,
    T_DECORSTR,
    T_DESTRIPE,
    T_DISK2IMG,
    T_DUPL,
    T_EDGE,
    T_EIGEN,
    T_EXPAND,
    T_FCONVL,
    T_FFT1D,
    T_FFT2D,
    T_FHT1D,
    T_FHT2D,
    T_FILTER,
    T_FINI,
    T_FLTBOX,
    T_FLTCONE,
    T_FLTCYL,
    T_FLTEXP,
    T_FLTGAUSS,
    T_FUNCTION,
    T_GEOMCOEF,
    T_GEOMWARP,
    T_GETMEM,
    T_GRADIENT,
    T_GRAYSCAL,
    T_HELP,
    T_HISTEQ,
    T_HISTOGRM,
    T_HMOSAIC,
    T_HSVTORGB,
    T_IMG2DISK,
    T_IMG2EPSF,
    T_IMG2TEXT,
    T_IMG2TIFF,
    T_IMGINFO,
    T_IMPORT,
    T_INIT,
    T_INSERT,
    T_INVERT,
    T_JPEG2IMG,
    T_LABEL,
    T_LVLSLICE,
    T_MAIN,
    T_MATCH,
    T_MAXLIKE,
    T_MEDIAN,
    T_MESSAGE,
    T_MINDIST,
    T_MOSAIC,
    T_PACK,
    T_PCT,
    T_PLOT,
    T_PROGRESS,
    T_PSPECT,
    T_PSPECT1D,
    T_RANDOM,
    T_RANDOM3D,
    T_RANGE,
    T_REAL2COM,
    T_REFINE,
    T_RELIEF,
    T_RELIMG,
    T_RELMEM,
    T_RESAMPL,
    T_RGBTOHSV,
    T_RMSDIFF,
    T_ROISTATS,
    T_ROTATE,
    T_ROTATECUBE,
    T_SCATTER,
    T_SCONVL,
    T_SEGMENT,
    T_SIGMAP,
    T_SINESTAR,
    T_SINEWAVE,
    T_STATS,
    T_STRETCH,
    T_SUBSAMPL,
    T_SWAP2,
    T_SWAP4,
    T_TEST,
    T_TFINVERT,
    T_TFLINEAR,
    T_TFLOG,
    T_TFPLT,
    T_TFPOWER,
    T_TFQUANT,
    T_TFROOT,
    T_TFSAT,
    T_TFSCALE,
    T_TFSQUARE,
    T_TFTABLE,
    T_TFTHRESH,
    T_TIFF2IMG,
    T_TIMING,
    T_TRNSFORM,
    T_UNPACK,
    T_VMOSAIC,
    T_VSTRETCH,
    T_WINDOW,
    T_WTFCAVG,
    T_WTFDIVER,
    T_WTFMAJ,
    T_WTFMAX,
    T_WTFMIN,
    T_WTFVAR,
    T_END
} T_TIME;

#ifdef SADIE_LIB

enum message_level {                    /* Severity levels for messages */
    SAD_ERROR,
    SAD_WARNING,
    SAD_INFO,
    SAD_DEFAULT,
    SAD_NLEVELS
};

/* Application message handling callback functions  */
typedef void (*app_messenger_t)(char option, char * msg, void * data);

/* Application histogram plotting callback functions  */
typedef int (*app_plotter_t)(PIXEL * line, uint32_t nbnd, uint32_t npix,
                             const char * minxlbl, const char * maxxlbl,
                             int option, PIXEL gmin, PIXEL gmax,
                             void * data);

/* Application progress meter creation callback functions */
typedef int (*app_progmeter_creator_t)(const char * metername, void * data);

/* Application progress meter updating callback functions */
typedef int (*app_progmeter_updator_t)(int meterid, double value, void * data);

/* Application progress meter destruction callback functions */
typedef int (*app_progmeter_destructor_t)(int meterid, void * data);

/* Application private data destructor callbacks */
typedef int (*app_data_destructor_t)(void * data);

extern int sad_register_message_callback (int level, app_messenger_t thecallback);
extern int sad_register_plotter_callback (app_plotter_t thecallback);
extern int sad_register_creator_callback (app_progmeter_creator_t thecallback);
extern int sad_register_updator_callback (app_progmeter_updator_t thecallback);
extern int sad_register_destructor_callback (app_progmeter_destructor_t thecallback);
extern int sad_register_data_destructor (app_data_destructor_t thecallback);

extern int sad_add_app_data (void * thedata);
extern int sad_remove_app_data (void);

extern const char* sad_version_string (void);

/* Global variables used in various parts of SADIE */
extern uint32_t nlev;
extern uint32_t csize;
extern double weight;
extern double power;
extern double *count;
extern PIXEL gain;
extern PIXEL bias;
extern PIXEL gmin;
extern PIXEL gmax;
extern PIXEL thresh;
extern PIXEL gbrk[2][4];
extern PIXEL *table;

#endif /* defined SADIE_LIB */

extern char *functions[T_END][2];

extern char *SLOG;                      /* Name of the system log-file */

extern char *Environment;
extern short Application;
extern short AppVersion;

extern short NAMES;                     /* Switch to control amount of output to system log-file */

extern short TIMES;                     /* Switch to enable/disable timing facility */

extern short LINES;                     /* Switch to enable/disable progress indicator */

extern short ABORT;                     /* Flag */

/*----------------------------------------------------------------------------*/

#define         FALSE           0
#define         TRUE            1

#define         OFF             0
#define         ON              1

#define         BEGIN           "/*-"
#define         END             "/*-"

#define         CODE            3

#define         BIL             0
#define         BSQ             1
#define         BIP             2

#define         REIMAG          0
#define         MAGPHA          1

#define         ADD             0
#define         SUBTRACT        1
#define         MULTIPLY        2
#define         DIVIDE          3
#define         MINIMUM         4
#define         MAXIMUM         5

#define         LT              0
#define         LE              1
#define         GT              2
#define         GE              3
#define         EQ              4
#define         NE              5
#define         AND             6
#define         OR              7
#define         XOR             8
#define         NOT             9

#define         DESMNE          0
#define         DESSTD          1
#define         DESCDF          2

#define         DESSNG          0
#define         DESAVG          1

#define         SOBEL           0
#define         SOBEL2          1
#define         SOBEL3          2
#define         ROBERTS         3
#define         ROBERTS2        4
#define         PREWITT         5

#define         PSF             0
#define         OTF             1

#define         FORWARD         (-1)
#define         INVERSE         1

#define         NOR_PSPECT      0
#define         ROW_PSPECT      1
#define         COL_PSPECT      2

#define         CONE            0
#define         NEGEXP          1
#define         GAUSS           2
#define         BOX             3
#define         CYLINDER        4
#define         HANNING         5
#define         HAMMING         6
#define         KAISER          7
#define         SINC            8

#define         MINTERMS        3
#define         MAXTERMS        6

#define         LSQ             0
#define         CDF             1

#define         CITYBLOCK       0
#define         EUCLIDEAN       1
#define         MAHALANOBIS     2

#define         UNIT            0
#define         NORM            1
#define         LOG             2

#define         IGNORE          0
#define         INCLUDE         1
#define         EXCLUDE         2

#define         UNIFORM         0
#define         GAUSSIAN        1

#define         FLIP0           0
#define         FLIPM45         1
#define         FLIPP45         2
#define         FLIP90          3
#define         ROTPM0          4
#define         ROTP90          5
#define         ROTM90          6
#define         ROT180          7

#define         PI              3.14159265358979323846

/*----------------------------------------------------------------------------*/

#define         min(x,y)        ((x) < (y)  ?  (x) : (y))
#define         max(x,y)        ((x) > (y)  ?  (x) : (y))

#define         rnd(x)          (long long)((x) > 0.0  ?  (x)+0.5 : (x)-0.5)

/*----------------------------------------------------------------------------*/

extern  void    BILTOBSQ (IMAGE *, IMAGE **);
extern  void    BITPLANE (IMAGE *, uint32_t, IMAGE **);
extern  void    BSQTOBIL (IMAGE **, uint32_t, IMAGE **);
extern  void    CHECKER  (uint32_t, uint32_t, uint32_t, uint32_t, IMAGE **);
extern  short   CHECKIMG (IMAGE *);
extern  void    CHIRP    (uint32_t, uint32_t, double, IMAGE **);
extern  void    CLUSTER  (IMAGE *, short, uint32_t, uint32_t, PIXEL, uint32_t, uint32_t, uint32_t, PIXEL, IMAGE **);
extern  void    COM2REAL (IMAGE *, short, IMAGE **, IMAGE **);
extern  void    COMBINE  (IMAGE *, IMAGE *, short, double, double, IMAGE **);
extern  void    COMPARE  (IMAGE *, IMAGE *, short, IMAGE **);
extern  void    CONTNORM (IMAGE *, PIXEL, double, uint32_t, IMAGE **);
extern  void    CONTOUR  (IMAGE *, short, double, IMAGE **);
extern  void    COVAR    (IMAGE *, uint32_t, double *, double *);
extern  double  CSPLINE  (double, double, double, double, double, double);
extern  void    DECORSTR (IMAGE *, uint32_t, IMAGE **);
extern  void    DESTRIPE (IMAGE *, short, short, short, short, short, IMAGE **);
extern  void    DISK2IMG (unsigned char *, short, uint64_t, IMAGE **);
extern  void    DUPL     (IMAGE *, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, IMAGE **);
extern  void    EDGE     (IMAGE *, short, IMAGE **, IMAGE **);
extern  void    EIGEN    (double *, uint32_t, double *);
extern  void    EXPAND   (IMAGE *, double, double, double, IMAGE **);
extern  void    FCONVL   (IMAGE *, IMAGE *, short, IMAGE **);
extern  void    FFT1D    (short, uint32_t, double *, double *, double *);
extern  void    FFT2D    (IMAGE *, short, IMAGE **);
extern  void    FHT1D    (short, uint32_t, uint32_t, uint32_t *, uint32_t *, double *, double *, double *);
extern  void    FHT2D    (IMAGE *, short, IMAGE **);
extern  void    FILTER   (IMAGE *, double (*)(double, double, double, double), double, double, double, IMAGE **);
extern  void    FINI     (void);
extern  double  FLTBOX   (double, double, double, double);
extern  double  FLTCONE  (double, double, double, double);
extern  double  FLTCYL   (double, double, double, double);
extern  double  FLTEXP   (double, double, double, double);
extern  double  FLTGAUSS (double, double, double, double);
extern  void    FUNCTION (uint32_t, uint32_t, double, double, double, short, IMAGE **);
extern  void    GEOMCOEF (double *, double *, double *, double *, short, short, double *, double *);
extern  void    GEOMWARP (IMAGE *, uint32_t, uint32_t, uint32_t, double *, double *, short, double, PIXEL, IMAGE **);
extern  void    GETMEM   (uint32_t, uint32_t, uint32_t, IMAGE **);
extern  void    GRADIENT (IMAGE *, PIXEL *, PIXEL *, short, IMAGE **, IMAGE **);
extern  void    GRAYSCAL (uint32_t, uint32_t, uint32_t, PIXEL, PIXEL, IMAGE **);
extern  void    HELP     (T_TIME, char *);
extern  void    HISTEQ   (IMAGE *, uint32_t, uint32_t, IMAGE **);
extern  void    HISTOGRM (IMAGE *, uint32_t, short, PIXEL, PIXEL, uint32_t, uint64_t *);
extern  void    HMOSAIC  (IMAGE *, IMAGE *, int32_t, PIXEL, IMAGE **);
extern  void    HSVTORGB (IMAGE *, IMAGE **);
extern  void    IMG2DISK (IMAGE *, unsigned char *, uint32_t, short, short);
extern  void    IMG2EPSF (IMAGE *, unsigned char *, uint32_t, short, short);
extern  void    IMG2TEXT (IMAGE *, unsigned char *, short, short);
extern  void    IMG2TIFF (IMAGE *, unsigned char *, uint32_t, short, short);
extern  void    IMPORT   (unsigned char *, short, short, uint64_t, uint64_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, IMAGE **);
extern  void    IMGINFO  (IMAGE *);
extern  void    INIT     (void);
extern  void    INSERT   (IMAGE *, IMAGE *, uint32_t, uint32_t, IMAGE **);
extern  void    INVERT   (double *, uint32_t, double *);
extern  void    JPEG2IMG (unsigned char *, short, IMAGE **);
extern  void    LABEL    (IMAGE *, double, PIXEL, IMAGE **);
extern  void    LVLSLICE (IMAGE *, PIXEL *, uint32_t, IMAGE **);
extern  void    MATCH    (IMAGE *, IMAGE *, uint32_t, short, IMAGE **);
extern  void    MAXLIKE  (IMAGE *, uint32_t, double *, double *, IMAGE **);                              
extern  void    MEDIAN   (IMAGE *, uint32_t, uint32_t, IMAGE **);
extern  void    MESSAGE  (char, char *);
extern  void    MINDIST  (IMAGE *, short, uint32_t, double *, double *, IMAGE **);                              
extern  void    MOSAIC   (IMAGE **, uint32_t, uint32_t, uint32_t, uint32_t *, uint32_t *, PIXEL, IMAGE **);
extern  void    PACK     (uint32_t, uint64_t, PIXEL *);
extern  void    PCT      (IMAGE *, uint32_t, IMAGE **);
extern  short   PLOT     (PIXEL *, uint32_t, uint32_t, char *, char *, short);
extern  short   PROGRESS (double);
extern  void    PSPECT   (IMAGE *, IMAGE **);
extern  void    PSPECT1D (IMAGE *, short, IMAGE **);
extern  void    RANDOM   (uint32_t, uint32_t, short, double, IMAGE **);
extern  void    RANDOM3D (uint32_t, uint32_t, uint32_t, short, double, IMAGE **);
extern  void    RANGE    (IMAGE *);
extern  void    REAL2COM (IMAGE *, IMAGE *, short, IMAGE **);
extern  void    REFINE   (IMAGE *[3], IMAGE **);
extern  void    RELIEF   (IMAGE *, double, double, double, IMAGE **);
extern  void    RELIMG   (IMAGE **);
extern  void    RELMEM   (IMAGE *);
extern  void    RESAMPL  (IMAGE *, uint32_t, uint32_t, uint32_t, uint32_t, IMAGE **);
extern  void    RGBTOHSV (IMAGE *, IMAGE **);
extern  void    RMSDIFF  (IMAGE *, IMAGE *, double *);
extern  void    ROISTATS (char *, uint32_t, uint64_t, PIXEL *, uint32_t, uint32_t, PIXEL *, PIXEL *, double *, double *, uint64_t *, double *, double *);
extern  void    ROTATE   (IMAGE *, double, PIXEL, double, IMAGE **);
extern  void    ROTATECUBE (IMAGE *, short, short, short, short, short, short, IMAGE **);
extern  void    SCATTER  (IMAGE *, IMAGE *, uint32_t, uint32_t, uint32_t, uint32_t, IMAGE **);
extern  void    SCONVL   (IMAGE *, PIXEL *, uint32_t, uint32_t, IMAGE **);
extern  void    SEGMENT  (IMAGE *, short, double, PIXEL, IMAGE **);
extern  void    SIGMAP   (IMAGE *, IMAGE *, IMAGE **);
extern  void    SINESTAR (uint32_t, uint32_t, double, IMAGE **);
extern  void    SINEWAVE (uint32_t, uint32_t, double, double, IMAGE **);
extern  void    STATS    (IMAGE *, uint32_t, uint32_t, short, PIXEL, PIXEL, double *, double *, PIXEL *, PIXEL *, PIXEL *, uint64_t *);
extern  void    STRETCH  (IMAGE *, void (*)(IMAGE *, IMAGE **), IMAGE **);
extern  void    SUBSAMPL (IMAGE *, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, uint32_t, IMAGE **);
extern  void    SWAP2    (BYTE *);
extern  void    SWAP4    (BYTE *);
extern  void    TFINVERT (IMAGE *, IMAGE **);
extern  void    TFLINEAR (IMAGE *, IMAGE **);
extern  void    TFLOG    (IMAGE *, IMAGE **);
extern  void    TFPLT    (IMAGE *, IMAGE **);
extern  void    TFPOWER  (IMAGE *, IMAGE **);
extern  void    TFQUANT  (IMAGE *, IMAGE **);
extern  void    TFROOT   (IMAGE *, IMAGE **);
extern  void    TFSAT    (IMAGE *, IMAGE **);
extern  void    TFSCALE  (IMAGE *, IMAGE **);
extern  void    TFSQUARE (IMAGE *, IMAGE **);
extern  void    TFTABLE  (IMAGE *, IMAGE **);
extern  void    TFTHRESH (IMAGE *, IMAGE **);
extern  void    TIFF2IMG (unsigned char *, short, IMAGE **);
extern  void    TIMING   (T_TIME);
extern  void    TRNSFORM (IMAGE *, short, IMAGE **);
extern  void    UNPACK   (uint32_t, uint64_t, PIXEL *);
extern  void    VMOSAIC  (IMAGE *, IMAGE *, int32_t, PIXEL, IMAGE **);
extern  void    VSTRETCH (IMAGE *, uint32_t, IMAGE **);
extern  void    WINDOW   (IMAGE *, PIXEL (*)(PIXEL *, uint32_t, uint32_t), uint32_t, uint32_t, IMAGE **);
extern  PIXEL   WTFCAVG  (PIXEL *, uint32_t, uint32_t);
extern  PIXEL   WTFDIVER (PIXEL *, uint32_t, uint32_t);
extern  PIXEL   WTFMAJ   (PIXEL *, uint32_t, uint32_t);
extern  PIXEL   WTFMAX   (PIXEL *, uint32_t, uint32_t);
extern  PIXEL   WTFMIN   (PIXEL *, uint32_t, uint32_t);
extern  PIXEL   WTFVAR   (PIXEL *, uint32_t, uint32_t);

/*----------------------------------------------------------------------------*/

#ifdef __cplusplus
}
#endif

#endif /* !defined SADIE_H */
