#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1990 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function optionally resamples an image, creates a label map of      */
/*   the image by segmenting it into regions of similar graylevels, and,      */
/*   if the input image was resampled, refines the label map back to the      */
/*   original resolution of the input image.                                  */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Background Information-----------------------------------------------------*/
/*                                                                            */
/*   Mehldau, G.:                                                             */
/*   "An Image Segmentation Algorithm."                                       */
/*   University of Arizona, DIAL, Technical Report No. 90-3                   */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void SEGMENT (
IMAGE  *in,     /*  I   Pointer to the input image.                           */
short  steps,   /*  I   Number of reduction steps.                            */
double sigma,   /*  I   Greylevel variance threshold.                         */
PIXEL  thresh,  /*  I   Greylevel difference threshold.                       */
IMAGE  **out    /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register short i;
    uint64_t total;
    char  msg[SLEN];
    IMAGE **scratch=NULL, *img[3];

    if (TIMES) TIMING(T_SEGMENT);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"SEGMENT");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (steps < 0) {
        MESSAGE('E'," Number of reduction steps must be equal to or greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                      %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Graylevel difference threshold: %12.4e",thresh);
        MESSAGE('I',msg);
        sprintf(msg," Graylevel variance threshold:   %12.4e",sigma);
        MESSAGE('I',msg);
        sprintf(msg," Number of reduction steps:        %d",steps);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* allocate and initialize image vector */
    total = ((uint64_t)2*(uint64_t)steps + (uint64_t)2) * (uint64_t)sizeof(IMAGE *);
    if (total > (uint64_t)SIZE_MAX  ||  (scratch=(IMAGE **)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    for (i=0; i<2*steps+2; i++) {
        scratch[i] = 0;
    }

    /* reduce image */
    scratch[0] = in;
    for (i=0; i<steps; i++) {
        RESAMPL(scratch[i],2,2,2,2,&scratch[i+1]);
        if (ABORT  ||  !scratch[i+1]) goto the_end;
        sprintf(scratch[i+1]->text,"Scratch");
    }

    /* label image */
    LABEL(scratch[steps],sigma,thresh,&scratch[steps+1]);
    if (ABORT  ||  !scratch[steps+1]) goto the_end;
    sprintf(scratch[steps+1]->text,"Scratch");

    /* refine image */
    for (i=0; i<steps; i++) {
        img[0] = scratch[steps+i+1];
        img[1] = scratch[steps-i];
        img[2] = scratch[steps-i-1];
        REFINE(img,&scratch[steps+i+2]);
        if (ABORT  ||  !scratch[steps+i+2]) goto the_end;
        sprintf(scratch[steps+i+2]->text,"Scratch");
        RELMEM(scratch[steps-i]);
        scratch[steps-i] = NULL;
    }
    *out = scratch[2*steps+1];
    
    the_end:
    if (scratch) {
        for (i=1; i<2*steps+1; i++) {
            if (scratch[i]) RELMEM(scratch[i]);
        }
        free(scratch);
    }
    if (TIMES) TIMING(T_EXIT);
}
