#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function computes the mean, standard deviation, minimum, maximum,   */
/*   and range of graylevels in an image, as well as the number of pixels     */
/*   with a zero graylevel, and outputs a histogram.                          */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void STATS (
IMAGE  *in,     /*  I   Pointer to the input image.                           */
uint32_t inc,   /*  I   Pixels/line and line increment.                       */
uint32_t nlev,  /*  I   > 0:   Number of bins in the histogram.               */
                /*      = 0:   Histogram is suppressed.                       */
short  option,  /*  I   Switch, indicating how to use input pixel range:      */
                /*      INCLUDE - Use only this input range for computing     */
                /*                statistics.                                 */
                /*      EXCLUDE - Exclude this pixel range.                   */
                /*      IGNORE  - Do not use the pixel range.                 */
PIXEL  rangemin,/*  I   Minimum pixel value in range of interest.             */
PIXEL  rangemax,/*  I   Maximum pixel value in range of interest.             */
double *mean,   /*  O   Pointer to an array[nbnd], containing the             */
                /*      mean graylevel for each band.                         */
double *dev,    /*  O   Pointer to an array[nbnd], containing the             */
                /*      graylevel standard deviation for each band.           */
PIXEL  *gmin,   /*  O   Pointer to an array[nbnd], containing the             */
                /*      minimum graylevel for each band.                      */
PIXEL  *gmax,   /*  O   Pointer to an array[nbnd], containing the             */
                /*      maximum graylevel for each band.                      */
PIXEL  *grng,   /*  O   Pointer to an array[nbnd], containing the             */
                /*      graylevel range for each band.                        */
uint64_t *zero  /*  O   Pointer to an array[nbnd], containing the             */
                /*      number of pixels with a zero graylevel for each band. */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k;
    uint64_t total, *include=NULL, *exclude=NULL, *hist=NULL;
    char msg[SLEN], minlbl[SLEN], maxlbl[SLEN];
    double sum, sumsq;
    PIXEL  *phist=NULL;

    if (TIMES) TIMING(T_STATS);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"STATS");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (inc <= 0) {
        MESSAGE('E'," Increment must be greater than zero.");
        goto the_end;
    } else if (nlev < 0) {
        MESSAGE('E'," Number of histogram bins must be greater than or equal to zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                         %s",in->text);
        MESSAGE('I',msg);
        sprintf(msg," Line and pixel subsample increment:  %lu",inc);
        MESSAGE('I',msg);
        if (nlev > 0) {
            sprintf(msg," Histogram resolution:                %lu",nlev);
            MESSAGE('I',msg);
        }
        if (option == INCLUDE) {
            sprintf(msg," Include only pixels in the graylevel range: %.2f to %.2f",rangemin,rangemax);
            MESSAGE('I',msg);
        } else if (option == EXCLUDE) {
            sprintf(msg," Exclude pixels in the graylevel range: %.2f to %.2f",rangemin,rangemax);
            MESSAGE('I',msg);
        }
        MESSAGE('I'," ...............");
    }

    /* allocate space */
    total = (uint64_t)in->nbnd*(uint64_t)sizeof(uint64_t);
    if (total > (uint64_t)SIZE_MAX  ||  (include=(uint64_t *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    total = (uint64_t)in->nbnd*(uint64_t)sizeof(uint64_t);
    if (total > (uint64_t)SIZE_MAX  ||  (exclude=(uint64_t *)malloc((size_t)total)) == NULL) {
        MESSAGE('E'," Memory request failed.");
        goto the_end;
    }
    if (nlev > 0) {
        total = (uint64_t)in->nbnd*(uint64_t)nlev*(uint64_t)sizeof(uint64_t);
        if (total > (uint64_t)SIZE_MAX  ||  (hist=(uint64_t *)malloc((size_t)total)) == NULL) {
            MESSAGE('E'," Memory request failed.");
            goto the_end;
        }
        total = (uint64_t)in->nbnd*(uint64_t)nlev*(uint64_t)sizeof(PIXEL);
        if (total > (uint64_t)SIZE_MAX  ||  (phist=(PIXEL *)malloc((size_t)total)) == NULL) {
            MESSAGE('E'," Memory request failed.");
            goto the_end;
        }
    }

    /* compute and output statistics for each band */
    MESSAGE('I',"");
    if (option == IGNORE) {
        MESSAGE('I'," Band     Minimum     Maximum        Mean   Deviation       Total      Zero/Non-Zero");
    } else if (option == INCLUDE  ||  option == EXCLUDE) {
        MESSAGE('I'," Band     Minimum     Maximum        Mean   Deviation       Total      Zero/Non-Zero   Included/Excluded");
    }
    MESSAGE('I',"");
    for (i=0; i<in->nbnd; i++) {
        gmin[i] = in->data[i][0][0];
        gmax[i] = in->data[i][0][0];
        zero[i] = 0;
        include[i] = 0;
        exclude[i] = 0;
        if (option == IGNORE) {
            for (sum=sumsq=0.0,j=0; j<in->nlin; j+=inc) {
                for (k=0; k<in->npix; k+=inc) {
                    sum    += (double)in->data[i][j][k];
                    sumsq  += (double)in->data[i][j][k]*(double)in->data[i][j][k];
                    gmin[i] = min(gmin[i],in->data[i][j][k]);
                    gmax[i] = max(gmax[i],in->data[i][j][k]);
                    if (in->data[i][j][k] == (PIXEL)0) {
                        zero[i] += 1;
                    }
                }
            }
            mean[i] = sum/((double)(in->nlin/inc+(in->nlin%inc ? 1:0))*(double)(in->npix/inc+(in->npix%inc ? 1:0)));
            dev[i]  = sqrt(sumsq/((double)(in->nlin/inc+(in->nlin%inc ? 1:0))*(double)(in->npix/inc+(in->npix%inc ? 1:0)))-mean[i]*mean[i]);
            grng[i] = gmax[i]-gmin[i];
            sprintf(msg,"%5lu%12.4e%12.4e%12.4e%12.4e%12llu%10llu/%llu",i+1,(double)gmin[i],(double)gmax[i],mean[i],dev[i],(uint64_t)in->nlin*(uint64_t)in->npix,zero[i],(uint64_t)in->nlin*(uint64_t)in->npix-zero[i]);
            MESSAGE('I',msg);
        } else if (option == INCLUDE) {
            for (sum=sumsq=0.0,j=0; j<in->nlin; j+=inc) {
                for (k=0; k<in->npix; k+=inc) {
                    if (in->data[i][j][k] >= rangemin && in->data[i][j][k] <= rangemax) {
                        sum    += (double)in->data[i][j][k];
                        sumsq  += (double)in->data[i][j][k]*(double)in->data[i][j][k];
                        gmin[i] = min(gmin[i],in->data[i][j][k]);
                        gmax[i] = max(gmax[i],in->data[i][j][k]);
                        if (in->data[i][j][k] == (PIXEL)0) {
                            zero[i] += 1;
                        }
                        include[i] += 1;
                    } else {
                        exclude[i] += 1;
                    }
                }
            }
            if (include[i] > 0) {
                mean[i] = sum/((double)include[i]);
                dev[i]  = sqrt(sumsq/((double)include[i])-mean[i]*mean[i]);
                grng[i] = gmax[i]-gmin[i];
            } else {
                mean[i] = 0.0;
                dev[i]  = 0.0;
                grng[i] = 0.0;
                gmin[i] = 0.0;
                gmax[i] = 0.0;
            }
            sprintf(msg,"%5lu%12.4e%12.4e%12.4e%12.4e%12llu%10llu/%llu%15llu/%llu",i+1,(double)gmin[i],(double)gmax[i],mean[i],dev[i],(uint64_t)in->nlin*(uint64_t)in->npix,zero[i],include[i]-zero[i],include[i],exclude[i]);
            MESSAGE('I',msg);
        } else if (option == EXCLUDE) {
            for (sum=sumsq=0.0,j=0; j<in->nlin; j+=inc) {
                for (k=0; k<in->npix; k+=inc) {
                    if (in->data[i][j][k] < rangemin || in->data[i][j][k] > rangemax) {
                        sum    += (double)in->data[i][j][k];
                        sumsq  += (double)in->data[i][j][k]*(double)in->data[i][j][k];
                        gmin[i] = min(gmin[i],in->data[i][j][k]);
                        gmax[i] = max(gmax[i],in->data[i][j][k]);
                        if (in->data[i][j][k] == (PIXEL)0) {
                            zero[i] += 1;
                        }
                        include[i] += 1;
                    } else {
                        exclude[i] += 1;
                    }
                }
            }
            if (include[i] > 0) {
                mean[i] = sum/((double)include[i]);
                dev[i]  = sqrt(sumsq/((double)include[i])-mean[i]*mean[i]);
                grng[i] = gmax[i]-gmin[i];
            } else {
                mean[i] = 0.0;
                dev[i]  = 0.0;
                grng[i] = 0.0;
                gmin[i] = 0.0;
                gmax[i] = 0.0;
            }
            sprintf(msg,"%5lu%12.4e%12.4e%12.4e%12.4e%12llu%10llu/%llu%15llu/%llu",i+1,(double)gmin[i],(double)gmax[i],mean[i],dev[i],(uint64_t)in->nlin*(uint64_t)in->npix,zero[i],include[i]-zero[i],include[i],exclude[i]);
            MESSAGE('I',msg);
        }
    }

    /* compute and output histograms */
    if (nlev > 0) {
        RANGE(in);
        MESSAGE('I',"");
        if (option == IGNORE) {
            sprintf(msg," Histogram has %lu levels between %.4e and %.4e",nlev,(double)in->gmin,(double)in->gmax);
            MESSAGE('I',msg);
            HISTOGRM(in,inc,option,in->gmin,in->gmax,nlev,hist);
            sprintf(minlbl,"%.2f",in->gmin);
            sprintf(maxlbl,"%.2f",in->gmax);
        } else if (option == INCLUDE) {
            sprintf(msg," Histogram has %lu levels between %.4e and %.4e",nlev,(double)rangemin,(double)rangemax);
            MESSAGE('I',msg);
            HISTOGRM(in,inc,option,rangemin,rangemax,nlev,hist);
            sprintf(minlbl,"%.2f",rangemin);
            sprintf(maxlbl,"%.2f",rangemax);
        } else if (option == EXCLUDE) {
            sprintf(msg," Histogram has %lu levels between %.4e and %.4e",nlev,(double)in->gmin,(double)in->gmax);
            MESSAGE('I',msg);
            HISTOGRM(in,inc,option,rangemin,rangemax,nlev,hist);
            sprintf(minlbl,"%.2f",in->gmin);
            sprintf(maxlbl,"%.2f",in->gmax);
        }       
        for (i=0; i<in->nbnd*nlev; i++) {
            phist[i] = (PIXEL)hist[i];
        }
        PLOT(phist,in->nbnd,nlev,minlbl,maxlbl,NORM);
    }

    the_end:
    if (include) free(include);
    if (exclude) free(exclude);
    if (hist)    free(hist);
    if (phist)   free(phist);
    if (TIMES)   TIMING(T_EXIT);
}
