#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function extracts a subimage from an image.                         */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void SUBSAMPL (
IMAGE *in,      /*  I   Pointer to the input image.                           */
uint32_t ibgn,  /*  I   Number of the first band to be sampled.               */
uint32_t jbgn,  /*  I   Number of the first line to be sampled.               */
uint32_t kbgn,  /*  I   Number of the first pixel to be sampled.              */
uint32_t iinc,  /*  I   Band increment during sampling.                       */
uint32_t jinc,  /*  I   Line increment during sampling.                       */
uint32_t kinc,  /*  I   Pixel increment during sampling.                      */
uint32_t isize, /*  I   Number of bands to be spanned.                        */
uint32_t jsize, /*  I   Number of lines to be spanned.                        */
uint32_t ksize, /*  I   Number of pixels/line to be spanned.                  */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, l, m, n, nbnd, nlin, npix;
    char msg[SLEN];

    if (TIMES) TIMING(T_SUBSAMPL);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"SUBSAMPL");
    }

    /* check input */
    if (!CHECKIMG(in)) {
        MESSAGE('E'," Can't identify image.");
        goto the_end;
    } else if (0 > ibgn  ||  ibgn+isize > in->nbnd) {
        MESSAGE('E'," Image to be sampled must be located completely inside input image in the band dimension.");
        goto the_end;
    } else if (0 > jbgn  ||  jbgn+jsize > in->nlin) {
        MESSAGE('E'," Image to be sampled must be located completely inside input image in the line dimension.");
        goto the_end;
    } else if (0 > kbgn  ||  kbgn+ksize > in->npix) {
        MESSAGE('E'," Image to be sampled must be located completely inside input image in the pixel dimension.");
        goto the_end;
    } else if (iinc <= 0) {
        MESSAGE('E'," Band subsample increment must be greater than zero.");
        goto the_end;
    } else if (jinc <= 0) {
        MESSAGE('E'," Line subsample increment must be greater than zero.");
        goto the_end;
    } else if (kinc <= 0) {
        MESSAGE('E'," Pixel subsample increment must be greater than zero.");
        goto the_end;
    } else if (isize <= 0) {
        MESSAGE('E'," Number of bands to be spanned must be greater than zero.");
        goto the_end;
    } else if (jsize <= 0) {
        MESSAGE('E'," Number of lines to be spanned must be greater than zero.");
        goto the_end;
    } else if (ksize <= 0) {
        MESSAGE('E'," Number of pixels/line to be spanned must be greater than zero.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Input image:                                %s",in->text);
        MESSAGE('I',msg);
#ifndef SADIE_IPT
        sprintf(msg," From:                band, line, pixel:     %lu, %lu, %lu",ibgn,jbgn,kbgn);
#else
        sprintf(msg," From:                band, line, pixel:     %lu, %lu, %lu",ibgn+1,jbgn+1,kbgn+1);
#endif
        MESSAGE('I',msg);
#ifndef SADIE_IPT
        sprintf(msg," To:                  band, line, pixel:     %lu, %lu, %lu",ibgn+isize-1,jbgn+jsize-1,kbgn+ksize-1);
#else
        sprintf(msg," To:                  band, line, pixel:     %lu, %lu, %lu",ibgn+isize,jbgn+jsize,kbgn+ksize);
#endif
        MESSAGE('I',msg);
        sprintf(msg," Subsample increment: bands, lines, pixels:  %lu, %lu, %lu",iinc,jinc,kinc);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    nbnd = (uint32_t)ceil((double)isize/(double)iinc);
    nlin = (uint32_t)ceil((double)jsize/(double)jinc);
    npix = (uint32_t)ceil((double)ksize/(double)kinc);
    if (!CHECKIMG(*out)) GETMEM(nbnd,nlin,npix,out);
    if (!*out) goto the_end;

    /* extract subimage */
    for (i=0,l=ibgn; i<nbnd && l<in->nbnd; i+=1,l+=iinc) {
        for (j=0,m=jbgn; j<nlin && m<in->nlin; j+=1,m+=jinc) {
            for (k=0,n=kbgn; k<npix && n<in->npix; k+=1,n+=kinc) {
                (*out)->data[i][j][k] = in->data[l][m][n];
            }
        }
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
