#ifdef HAVE_CONFIG_H
#include <config.h>
#endif /* HAVE_CONFIG_H */
#include        "sadie.h"

/*-Copyright Information------------------------------------------------------*/
/* Copyright (c) 1988 by the University of Arizona Digital Image Analysis Lab */
/*----------------------------------------------------------------------------*/
/*-General Information--------------------------------------------------------*/
/*                                                                            */
/*   This function mosaics two images vertically.                             */
/*                                                                            */
/*----------------------------------------------------------------------------*/
/*-Interface Information------------------------------------------------------*/
void VMOSAIC (
IMAGE *in1,     /*  I   Pointer to the top input image.                       */
IMAGE *in2,     /*  I   Pointer to the bottom input image.                    */
int32_t kbgn,   /*  I   Number of columns (positive or negative) to offset    */
                /*      bottom image with respect to top image.               */
PIXEL glev,     /*  I   Graylevel to fill otherwise empty pixels.             */
IMAGE **out     /*  O   Address of a pointer to the output image.             */
/*----------------------------------------------------------------------------*/
) { register uint32_t i, j, k, nbnd, nlin, npix;
    char msg[SLEN];

    if (TIMES) TIMING(T_VMOSAIC);
    if (NAMES) {
        MESSAGE('I',"");
        MESSAGE('I',"VMOSAIC");
    }

    /* check input */
    if (!CHECKIMG(in1)) {
        MESSAGE('E'," Can't identify upper image.");
        goto the_end;
    } else if (!CHECKIMG(in2)) {
        MESSAGE('E'," Can't identify lower image.");
        goto the_end;
    } else if (in1->nbnd != in2->nbnd) {
        MESSAGE('E'," Number of bands must be identical in both images.");
        goto the_end;
    } else if ((uint64_t)in1->nlin+(uint64_t)in2->nlin > (uint64_t)UINT32_MAX) {
        MESSAGE('E'," Image size (number of lines) too large.");
        goto the_end;
    } else if (kbgn > 0  &&  (uint64_t)in2->npix+(uint64_t)kbgn > (uint64_t)UINT32_MAX) {
        MESSAGE('E'," Image size (number of pixels/line) too large.");
        goto the_end;
    } else if (kbgn <= 0  &&  (uint64_t)in1->npix-(uint64_t)kbgn > (uint64_t)UINT32_MAX) {
        MESSAGE('E'," Image size (number of pixels/line) too large.");
        goto the_end;
    }

    if (NAMES) {
        MESSAGE('I',"");
        sprintf(msg," Top input image:      %s",in1->text);
        MESSAGE('I',msg);
        sprintf(msg," Bottom input image:   %s",in2->text);
        MESSAGE('I',msg);
        sprintf(msg," Horizontal offset:    %ld",kbgn);
        MESSAGE('I',msg); 
        sprintf(msg," Fill graylevel:     %12.4e",glev);
        MESSAGE('I',msg);
        MESSAGE('I'," ...............");
    }

    /* create image of appropriate size */
    nbnd = in1->nbnd;
    nlin = in1->nlin + in2->nlin;
    npix = kbgn > 0  ?  max(in1->npix,in2->npix+kbgn) : max(in1->npix-kbgn,in2->npix);
    if (!CHECKIMG(*out)) GETMEM(nbnd,nlin,npix,out);
    if (!*out) goto the_end;

    /* concatenate images */
    for (i=0; i<nbnd; i++) {
        for (j=0; j<nlin; j++) {
            for (k=0; k<npix; k++) {
                (*out)->data[i][j][k] = glev;
            }
        }
        if (kbgn > 0) {
            for (j=0; j<in1->nlin; j++) {
                for (k=0; k<in1->npix; k++) {
                    (*out)->data[i][j][k] = in1->data[i][j][k];
                }
            }
            for (j=0; j<in2->nlin; j++) {
                for (k=0; k<in2->npix; k++) {
                    (*out)->data[i][j+in1->nlin][k+kbgn] = in2->data[i][j][k];
                }
            }
        } else /* (kbgn <= 0) */ {
            for (j=0; j<in1->nlin; j++) {
                for (k=0; k<in1->npix; k++) {
                    (*out)->data[i][j][k-kbgn] = in1->data[i][j][k];
                }
            }
            for (j=0; j<in2->nlin; j++) {
                for (k=0; k<in2->npix; k++) {
                    (*out)->data[i][j+in1->nlin][k] = in2->data[i][j][k];
                }
            }
        }
    }

    /* compute new minimum and maximum */
    if (in1->npix == in2->npix  &&  kbgn == 0) {
        (*out)->gmin = min(in1->gmin,in2->gmin);
        (*out)->gmax = max(in1->gmax,in2->gmax);
    } else {
        (*out)->gmin = min(glev,min(in1->gmin,in2->gmin));
        (*out)->gmax = max(glev,max(in1->gmax,in2->gmax));
    }

    the_end:
    if (TIMES) TIMING(T_EXIT);
}
